#!/usr/bin/env python3

"""
Tests for signalslot wrappers
"""

from ..nosecompat import eq_, assert_set_equal

from aioax25.signal import Slot

def test_slot_call():
    """
    Test the __call__ method passes the arguments given.
    """
    calls = []
    slot = Slot(lambda **kwa : calls.append(kwa),
            constkeyword1=1, constkeyword2=2, keyword3=3)

    slot(callkeyword1=1, callkeyword2=2, keyword3=30)

    # We should have a single call
    eq_(len(calls), 1)
    kwargs = calls.pop(0)

    # kwargs should contain the merged keywords
    assert_set_equal(set(kwargs.keys()), set([
        'constkeyword1', 'constkeyword2', 'callkeyword1', 'callkeyword2',
        'keyword3']))
    eq_(kwargs['constkeyword1'], 1)
    eq_(kwargs['constkeyword2'], 2)
    eq_(kwargs['callkeyword1'], 1)
    eq_(kwargs['callkeyword2'], 2)

    # This is given by both, call should override constructor
    eq_(kwargs['keyword3'], 30)

def test_slot_exception():
    """
    Test the __call__ swallows exceptions.
    """
    def _slot(**kwargs):
        raise RuntimeError('Whoopsie')
    slot = Slot(_slot)

    # This should not trigger an exception
    slot()



#!/usr/bin/env python3

from aioax25.signal import Signal
from aioax25.interface import AX25Interface
from aioax25.frame import AX25UnnumberedInformationFrame

from ..asynctest import asynctest
from asyncio import Future, get_event_loop, sleep

from ..nosecompat import assert_greater, assert_less, assert_is, \
        assert_greater_equal, eq_

import time
import re


class DummyKISS(object):
    """
    Dummy KISS interface for unit testing.
    """
    def __init__(self):
        self.received = Signal()
        self.sent = []

    def send(self, frame):
        self.sent.append((time.monotonic(), frame))


class UnreliableDummyKISS(DummyKISS):
    def __init__(self):
        super(UnreliableDummyKISS, self).__init__()
        self.send_calls = 0

    def send(self, frame):
        self.send_calls += 1
        if self.send_calls == 1:
            raise IOError('Whoopsie')
        super(UnreliableDummyKISS, self).send(frame)



@asynctest
async def test_received_msg_signal():
    """
    Test received messages trigger the received_msg signal.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    receive_future = Future()

    my_interface = AX25Interface(my_port)

    def _on_receive_match(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            receive_future.set_result(None)
        except Exception as e:
            receive_future.set_exception(e)
    my_interface.received_msg.connect(_on_receive_match)

    # Pass in a message
    my_port.received.emit(frame=bytes(my_frame))

    await receive_future


def test_receive_bind():
    """
    Test bind rejects non-strings as call-signs.
    """
    my_interface = AX25Interface(DummyKISS())
    try:
        my_interface.bind(
                callback=lambda *a, **kwa : None,
                callsign=123456,
                ssid=0,
                regex=False
        )
        assert False, 'This should not have worked'
    except TypeError as e:
        eq_(str(e), 'callsign must be a string (use '\
                    'regex=True for regex)')

@asynctest
async def test_receive_str_filter():
    """
    Test matching messages can trigger string filters (without SSID).
    """
    my_port = DummyKISS()
    unmatched_filter_received = []
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    receive_future = Future()

    my_interface = AX25Interface(my_port)

    def _on_receive_match(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            receive_future.set_result(None)
        except Exception as e:
            receive_future.set_exception(e)

    def _on_receive_nomatch(**kwargs):
        unmatched_filter_received.append(kwargs)

    def _on_timeout():
        receive_future.set_exception(AssertionError('Timed out'))

    # This should match
    my_interface.bind(_on_receive_match, 'VK4BWI', ssid=None)

    # This should not match
    my_interface.bind(_on_receive_nomatch, 'VK4AWI', ssid=None)

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Pass in a message
    my_port.received.emit(frame=bytes(my_frame))

    await receive_future
    eq_(len(unmatched_filter_received), 0)


@asynctest
async def test_receive_str_filter_ssid():
    """
    Test matching messages can trigger string filters (with SSID).
    """
    my_port = DummyKISS()
    unmatched_filter_received = []
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    receive_future = Future()

    my_interface = AX25Interface(my_port)

    def _on_receive_match(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            receive_future.set_result(None)
        except Exception as e:
            receive_future.set_exception(e)

    def _on_receive_nomatch(**kwargs):
        unmatched_filter_received.append(kwargs)

    def _on_timeout():
        receive_future.set_exception(AssertionError('Timed out'))

    # This should match
    my_interface.bind(_on_receive_match, 'VK4BWI', ssid=4)

    # This should not match
    my_interface.bind(_on_receive_nomatch, 'VK4BWI', ssid=3)

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Pass in a message
    my_port.received.emit(frame=bytes(my_frame))

    await receive_future
    eq_(len(unmatched_filter_received), 0)


@asynctest
async def test_receive_re_filter():
    """
    Test matching messages can trigger regex filters (without SSID).
    """
    my_port = DummyKISS()
    unmatched_filter_received = []
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    receive_future = Future()

    my_interface = AX25Interface(my_port)

    def _on_receive_match(interface, frame, match, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            receive_future.set_result(None)
        except Exception as e:
            receive_future.set_exception(e)

    def _on_receive_nomatch(**kwargs):
        unmatched_filter_received.append(kwargs)

    def _on_timeout():
        receive_future.set_exception(AssertionError('Timed out'))

    # This should match
    my_interface.bind(_on_receive_match, r'^VK4[BR]WI$',
            ssid=None, regex=True)

    # This should not match
    my_interface.bind(_on_receive_nomatch, r'^VK4[AZ]WI$',
            ssid=None, regex=True)

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Pass in a message
    my_port.received.emit(frame=bytes(my_frame))

    await receive_future
    eq_(len(unmatched_filter_received), 0)


@asynctest
async def test_receive_re_filter_ssid():
    """
    Test matching messages can trigger regex filters (with SSID).
    """
    my_port = DummyKISS()
    unmatched_filter_received = []
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    receive_future = Future()

    my_interface = AX25Interface(my_port)

    def _on_receive_match(interface, frame, match, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            receive_future.set_result(None)
        except Exception as e:
            receive_future.set_exception(e)

    def _on_receive_nomatch(**kwargs):
        unmatched_filter_received.append(kwargs)

    def _on_timeout():
        receive_future.set_exception(AssertionError('Timed out'))

    # This should match
    my_interface.bind(_on_receive_match, r'^VK4[BR]WI$', ssid=4, regex=True)

    # This should not match
    my_interface.bind(_on_receive_nomatch, r'^VK4[AZ]WI$', ssid=4, regex=True)

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Pass in a message
    my_port.received.emit(frame=bytes(my_frame))

    await receive_future
    eq_(len(unmatched_filter_received), 0)


def test_unbind_notexist_call():
    """
    Test unbinding a receiver for a call that does not exist returns silently.
    """
    my_interface = AX25Interface(DummyKISS())
    my_receiver = lambda **k : None

    # This should generate no error
    my_interface.unbind(my_receiver, 'MYCALL', ssid=12)

def test_unbind_notexist_ssid():
    """
    Test unbinding a receiver for a SSID that does not exist returns silently.
    """
    my_port = DummyKISS()
    my_interface = AX25Interface(my_port)

    my_receiver = lambda **k : None

    # Inject a receiver
    my_interface._receiver_str = {
            'MYCALL': {
                12: [
                    my_receiver
                ]
            }
    }

    # This should generate no error
    my_interface.unbind(my_receiver, 'MYCALL', ssid=14)

def test_unbind_notexist_receiver():
    """
    Test unbinding a receiver that is not bound should not raise error.
    """
    my_port = DummyKISS()
    my_interface = AX25Interface(my_port)

    my_receiver1 = lambda **k : None
    my_receiver2 = lambda **k : None

    # Inject a receiver
    my_interface._receiver_str = {
            'MYCALL': {
                12: [
                    my_receiver1
                ]
            }
    }

    # This should generate no error
    my_interface.unbind(my_receiver2, 'MYCALL', ssid=12)

def test_unbind_str():
    """
    Test unbinding a string receiver removes the receiver and cleans up.
    """
    my_port = DummyKISS()
    my_interface = AX25Interface(my_port)

    my_receiver = lambda **k : None

    # Inject a receiver
    my_interface._receiver_str = {
            'MYCALL': {
                12: [
                    my_receiver
                ]
            }
    }
    my_interface.unbind(my_receiver, 'MYCALL', ssid=12)

    # This should now be empty
    eq_(len(my_interface._receiver_str), 0)

def test_unbind_re():
    """
    Test unbinding a regex receiver removes the receiver and cleans up.
    """
    my_port = DummyKISS()
    my_interface = AX25Interface(my_port)

    my_receiver = lambda **k : None

    # Inject a receiver
    my_interface._receiver_re = {
            r'^MY': (re.compile(r'^MY'), {
                12: [
                    my_receiver
                ]
            })
    }
    my_interface.unbind(my_receiver, r'^MY', ssid=12, regex=True)

    # This should now be empty
    eq_(len(my_interface._receiver_re), 0)


def test_reception_resets_cts():
    """
    Check the clear-to-send expiry is updated with received traffic.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')

    my_interface = AX25Interface(my_port)
    cts_before = my_interface._cts_expiry

    # Pass in a message
    my_port.received.emit(frame=bytes(my_frame))
    cts_after = my_interface._cts_expiry

    assert_less(cts_before, cts_after)
    assert_greater(cts_after, time.monotonic())

@asynctest
async def test_transmit_waits_cts():
    """
    Test sending a message waits for the channel to be clear.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    transmit_future = Future()

    my_interface = AX25Interface(my_port, cts_delay=0.250)

    def _on_transmit(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            transmit_future.set_result(None)
        except Exception as e:
            transmit_future.set_exception(e)

    def _on_timeout():
        transmit_future.set_exception(AssertionError('Timed out'))

    # The time before transmission
    time_before = time.monotonic()

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Send the message
    my_interface.transmit(my_frame, _on_transmit)

    await transmit_future

    eq_(len(my_port.sent), 1)
    (send_time, sent_frame) = my_port.sent.pop(0)

    eq_(bytes(sent_frame), bytes(my_frame))
    assert_less((time.monotonic() - send_time), 0.05)
    assert_greater_equal((send_time - time_before), 0.25)


@asynctest
async def test_transmit_cancel():
    """
    Test that pending messages can be cancelled.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')

    my_interface = AX25Interface(my_port)

    # Send the message
    my_interface.transmit(my_frame)

    # Cancel it!
    my_interface.cancel_transmit(my_frame)

    # Wait a second
    await sleep(1)

    # Nothing should have been sent.
    eq_(len(my_port.sent), 0)


@asynctest
async def test_transmit_sends_immediate_if_cts():
    """
    Test the interface sends immediately if last activity a long time ago.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    transmit_future = Future()

    my_interface = AX25Interface(my_port)

    # Override clear to send expiry
    my_interface._cts_expiry = 0

    def _on_transmit(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            transmit_future.set_result(None)
        except Exception as e:
            transmit_future.set_exception(e)

    def _on_timeout():
        transmit_future.set_exception(AssertionError('Timed out'))

    # The time before transmission
    time_before = time.monotonic()

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Send the message
    my_interface.transmit(my_frame, _on_transmit)

    await transmit_future

    eq_(len(my_port.sent), 1)
    (send_time, sent_frame) = my_port.sent.pop(0)

    eq_(bytes(sent_frame), bytes(my_frame))
    assert_less((time.monotonic() - send_time), 0.05)
    assert_less((send_time - time_before), 0.01)


@asynctest
async def test_transmit_sends_if_not_expired():
    """
    Test the interface sends frame if not expired.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    my_frame.deadline = time.time() + 3600.0
    transmit_future = Future()

    my_interface = AX25Interface(my_port)

    # Override clear to send expiry
    my_interface._cts_expiry = 0

    def _on_transmit(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            transmit_future.set_result(None)
        except Exception as e:
            transmit_future.set_exception(e)

    def _on_timeout():
        transmit_future.set_exception(AssertionError('Timed out'))

    # The time before transmission
    time_before = time.monotonic()

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Send the message
    my_interface.transmit(my_frame, _on_transmit)

    await transmit_future

    eq_(len(my_port.sent), 1)
    (send_time, sent_frame) = my_port.sent.pop(0)

    eq_(bytes(sent_frame), bytes(my_frame))
    assert_less((time.monotonic() - send_time), 0.05)
    assert_less((send_time - time_before), 0.05)


@asynctest
async def test_transmit_drops_expired():
    """
    Test the interface drops expired messages.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    # This timestamp was a _long_ time ago!  1AM 1st January 1970
    my_frame.deadline = 3600
    transmit_future = Future()

    my_interface = AX25Interface(my_port)

    # Override clear to send expiry
    my_interface._cts_expiry = 0

    def _on_timeout():
        transmit_future.set_result(None)

    # The time before transmission
    time_before = time.monotonic()

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Send the message
    my_interface.transmit(my_frame)

    await transmit_future

    # Nothing should be sent!
    eq_(len(my_port.sent), 0)


@asynctest
async def test_transmit_waits_if_cts_reset():
    """
    Test the interface waits if CTS timer is reset.
    """
    my_port = DummyKISS()
    my_frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing')
    transmit_future = Future()

    my_interface = AX25Interface(my_port, cts_delay=0.250)

    def _on_transmit(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame), msg='Wrong frame')
            transmit_future.set_result(None)
        except Exception as e:
            transmit_future.set_exception(e)

    def _on_timeout():
        transmit_future.set_exception(AssertionError('Timed out'))

    # The time before transmission
    time_before = time.monotonic()

    # Set a timeout
    get_event_loop().call_later(1.0, _on_timeout)

    # Send the message
    my_interface.transmit(my_frame, _on_transmit)

    # Whilst that is pending, call reset_cts, this should delay transmission
    my_interface._reset_cts()

    await transmit_future

    eq_(len(my_port.sent), 1)
    (send_time, sent_frame) = my_port.sent.pop(0)

    eq_(bytes(sent_frame), bytes(my_frame))
    assert_less((time.monotonic() - send_time), 0.05)
    assert_greater(send_time - time_before, 0.25)
    assert_less(send_time - time_before, 1.05)


@asynctest
async def test_transmit_handles_failure():
    """
    Test transmit failures don't kill the interface handling.
    """
    my_port = UnreliableDummyKISS()
    my_frame_1 = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing 1')
    my_frame_2 = AX25UnnumberedInformationFrame(
            destination='VK4BWI-4',
            source='VK4MSL',
            pid=0xf0,
            payload=b'testing 2')
    transmit_future = Future()

    my_interface = AX25Interface(my_port, cts_delay=0.250)

    # Override clear to send expiry
    my_interface._cts_expiry = 0

    def _on_transmit(interface, frame, **kwargs):
        try:
            eq_(len(kwargs), 0, msg='Too many arguments')
            assert_is(interface, my_interface, msg='Wrong interface')
            eq_(bytes(frame), bytes(my_frame_2), msg='Wrong frame')
            transmit_future.set_result(None)
        except Exception as e:
            transmit_future.set_exception(e)

    def _on_timeout():
        transmit_future.set_exception(AssertionError('Timed out'))

    # The time before transmission
    time_before = time.monotonic()

    # Set a timeout
    get_event_loop().call_later(2.0, _on_timeout)

    # Send the messages
    my_interface.transmit(my_frame_1, _on_transmit) # This will fail
    my_interface.transmit(my_frame_2, _on_transmit) # This will work

    await transmit_future

    eq_(len(my_port.sent), 1)
    (send_time, sent_frame) = my_port.sent.pop(0)

    eq_(bytes(sent_frame), bytes(my_frame_2))
    assert_less((time.monotonic() - send_time), 0.05)
    assert_greater_equal((send_time - time_before), 0.25)



#!/usr/bin/env python3

"""
Asynchronous test case handler.
"""

from asyncio import get_event_loop
from functools import wraps

def asynctest(testcase, *args, **kwargs):
    """
    Wrap an asynchronous test case.
    """
    @wraps(testcase)
    def fn():
        return get_event_loop().run_until_complete(testcase(*args, **kwargs))

    return fn



#!/usr/bin/env python3

"""
Serial KISS interface unit tests.
"""

from collections import namedtuple
from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE
from aioax25 import kiss
import logging
from ..asynctest import asynctest
from asyncio import get_event_loop, sleep


class DummySerial(object):
    def __init__(self, port, baudrate, bytesize, parity, stopbits,
            timeout, xonxoff, rtscts, write_timeout, dsrdtr,
            inter_byte_timeout):

        assert port == '/dev/ttyS0'
        assert baudrate == 9600
        assert bytesize == EIGHTBITS
        assert parity == PARITY_NONE
        assert stopbits == STOPBITS_ONE
        assert timeout == None
        assert xonxoff == False
        assert rtscts == False
        assert write_timeout == None
        assert dsrdtr == False
        assert inter_byte_timeout == None

        self.rx_buffer = bytearray()
        self.tx_buffer = bytearray()
        self.flushes = 0
        self.closed = False
        self.read_exception = None

    def flush(self):
        self.flushes += 1

    def write(self, data):
        self.tx_buffer += data

    def read(self, length):
        if self.read_exception is not None:
            raise self.read_exception

        data = self.rx_buffer[0:length]
        self.rx_buffer = self.rx_buffer[length:]
        return data

    def close(self):
        self.closed = True

    @property
    def in_waiting(self):
        return len(self.rx_buffer)


# Dummy transport
class DummyTransport(object):
    def __init__(self, loop, port):
        self._loop = loop
        self._port = port

    def flush(self):
        future = self._loop.create_future()
        self._loop.call_soon(lambda : future.set_result(None))
        return future

    def write(self, *args, **kwargs):
        future = self._loop.create_future()
        self._port.write(*args, **kwargs)
        self._loop.call_soon(lambda : future.set_result(None))
        return future


# Keep a record of all ports, transports and protocols
PortConnection = namedtuple('PortConnection', ['port', 'protocol', 'transport'])
create_serial_conn_log = logging.getLogger('create_serial_connection')
connections = []

# Stub the serial port connection factory
async def dummy_create_serial_connection(loop, proto_factory, *args, **kwargs):
    future = loop.create_future()
    create_serial_conn_log.debug(
            'Creating new serial connection: '
            'loop=%r proto=%r args=%r kwargs=%r',
            loop, proto_factory, args, kwargs
    )

    def _open():
        create_serial_conn_log.debug('Creating objects')
        # Create the objects
        protocol = proto_factory()
        port = DummySerial(*args, **kwargs)
        transport = DummyTransport(loop, port)

        # Record the created object references
        connections.append(PortConnection(
            port=port, protocol=protocol, transport=transport
        ))

        # Pass the protocol the transport object
        create_serial_conn_log.debug('Passing transport to protocol')
        protocol.connection_made(transport)

        # Finish up the future
        create_serial_conn_log.debug('Finishing up')
        future.set_result((protocol, transport))

    create_serial_conn_log.debug('Scheduled in IOLoop')
    loop.call_soon(_open)

    create_serial_conn_log.debug('Returning future')
    return (await future)

kiss.create_serial_connection = dummy_create_serial_connection


class TestDevice(kiss.SerialKISSDevice):
    def __init__(self, *args, **kwargs):
        super(TestDevice, self).__init__(*args, **kwargs)
        self.init_called = False

    def _init_kiss(self):
        self.init_called = True


@asynctest
async def test_open():
    """
    Test we can open the port.
    """
    loop = get_event_loop()
    kissdev = TestDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop
    )
    assert kissdev._transport is None

    kissdev.open()
    await sleep(0.01)

    # We should have created a new port
    assert len(connections) == 1
    connection = connections.pop(0)

    # We should have a reference to the transport created.
    assert kissdev._transport == connection.transport

    # The device should have been initialised
    assert kissdev.init_called


@asynctest
async def test_close():
    """
    Test we can close the port.
    """
    loop = get_event_loop()
    kissdev = TestDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop, reset_on_close=False
    )

    # Force the port open
    kissdev._state = kiss.KISSDeviceState.OPEN
    serial = DummySerial(port='/dev/ttyS0', baudrate=9600,
            bytesize=EIGHTBITS, parity=PARITY_NONE, stopbits=STOPBITS_ONE,
            timeout=None, xonxoff=False, rtscts=False, write_timeout=None,
            dsrdtr=False, inter_byte_timeout=None)
    kissdev._transport = serial

    # Now try closing the port
    kissdev.close()
    await sleep(0.01)

    # The port should have been flushed
    assert serial.flushes == 1

    # The port should be closed
    assert serial.closed == True

    # The device should not reference the port
    assert kissdev._transport == None

    # The port should now be in the closed state
    assert kissdev._state == kiss.KISSDeviceState.CLOSED


def test_on_close_err(logger):
    """
    Test errors are logged if given
    """
    # Yeah, kludgy… but py.test won't see the fixture if I don't
    # do it this way.
    @asynctest
    async def _run():
        loop = get_event_loop()
        kissdev = TestDevice(
                device='/dev/ttyS0', baudrate=9600,
                log=logger, loop=loop, reset_on_close=False
        )

        # Force the port open
        kissdev._state = kiss.KISSDeviceState.OPEN
        serial = DummySerial(port='/dev/ttyS0', baudrate=9600,
                bytesize=EIGHTBITS, parity=PARITY_NONE, stopbits=STOPBITS_ONE,
                timeout=None, xonxoff=False, rtscts=False, write_timeout=None,
                dsrdtr=False, inter_byte_timeout=None)
        kissdev._transport = serial

        # Define a close error
        class CommsError(IOError):
            pass
        my_err = CommsError()

        # Now report the closure of the port
        kissdev._on_close(my_err)

        # We should have seen a log message reported
        assert logger.logrecords == [
                dict(
                    method='error',
                    args=('Closing port due to error %r', my_err,),
                    kwargs={},
                    ex_type=None, ex_val=None, ex_tb=None
                )
        ]

        # The device should not reference the port
        assert kissdev._transport == None

        # The port should now be in the closed state
        assert kissdev._state == kiss.KISSDeviceState.CLOSED
    _run()


@asynctest
async def test_send_raw_data():
    """
    Test _send_raw_data passes the data to the serial device.
    """
    loop = get_event_loop()
    kissdev = TestDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop
    )
    assert kissdev._transport is None

    kissdev.open()
    await sleep(0.01)

    # We should have created a new port
    assert len(connections) == 1
    connection = connections.pop(0)

    kissdev._send_raw_data(b'a test frame')
    assert bytes(connection.port.tx_buffer) == b'a test frame'



#!/usr/bin/env python3

"""
KISS device factory unit tests.
"""

from aioax25.kiss import \
        SerialKISSDevice, \
        SubprocKISSDevice, \
        TCPKISSDevice, \
        make_device

def test_serial_kiss():
    """
    Test we can create a ``SerialKISSDevice``.
    """
    dev = make_device(type='serial', device='/dev/ttyS0', baudrate=9600)
    assert isinstance(dev, SerialKISSDevice)
    assert dev._device == '/dev/ttyS0'
    assert dev._baudrate == 9600


def test_subproc_kiss():
    """
    Test we can create a ``SubprocKISSDevice``.
    """
    dev = make_device(type='subproc', command=['somecmd', 'a', 'b', 'c'])
    assert isinstance(dev, SubprocKISSDevice)
    assert dev._command == ['somecmd', 'a', 'b', 'c']
    assert dev._shell is False


def test_tcp_kiss():
    """
    Test we can create a ``TCPKISSDevice``.
    """
    dev = make_device(type='tcp', host='localhost', port=10000)
    assert isinstance(dev, TCPKISSDevice)
    assert dev._conn_args == dict(
            host='localhost', port=10000,
            ssl=None, family=0, proto=0, flags=0,
            sock=None, local_addr=None, server_hostname=None
    )


def test_unknown_kiss_type():
    """
    Test that unknown KISS device types are rejected
    """
    try:
        dev = make_device(type='bogus')
        assert False, 'Should not have worked, got a %r' % (dev,)
    except ValueError as e:
        assert str(e) == ("Unrecognised type=%r" % ("bogus",))



#!/usr/bin/env python3

"""
APRS position reporting handler.
"""

import math
from enum import Enum, IntEnum

from .frame import APRSFrame
from .datatype import APRSDataType
from .symbol import APRSSymbol
from .datetime import decode as decode_datetime
from .compression import compress, decompress, BYTE_VALUE_OFFSET
from ..unit import convertvalue, Quantity


class APRSPositionAmbiguity(Enum):
    NONE                = 0
    TENTH_MINUTE        = 1
    MINUTE              = 2
    TEN_MINUTES         = 3
    DEGREE              = 4


class APRSSexagesimal(object):
    """
    Represents a sexagesimal co-ordinate in degrees, minutes and seconds.
    """
    POS_SUFFIX = "+"
    NEG_SUFFIX = "-"
    STR_FORMAT = "%03d%02d.%02d%s"
    LENGTH = 9

    @classmethod
    def decode(cls, posstr):
        if len(posstr) < cls.LENGTH:
            raise ValueError("Position string too short")

        # Throw out any extra data
        posstr = posstr[0:cls.LENGTH]

        # Determine the sign
        signchr = posstr[-1]
        if signchr == cls.POS_SUFFIX:
            sign = 1
        elif signchr == cls.NEG_SUFFIX:
            sign = -1
        else:
            raise ValueError("Unrecognised sign: %r" % signchr)

        # There _must_ be a decimal point as the 3rd last character
        if posstr[-4] != ".":
            raise ValueError("No decimal point found in required position")

        # Determine the ambiguity
        digits = ''.join(reversed(posstr[-6:-4] + posstr[-3:-1]))
        try:
            ambiguity = digits.rindex(" ") + 1
        except ValueError:
            ambiguity = 0
        ambiguity = APRSPositionAmbiguity(ambiguity)

        # Decode the minutes
        minutes = float(posstr[-6:-1].replace(" ","0"))

        # The rest is the degrees
        degrees = int(posstr[0:-6])

        return cls(sign * degrees, minutes, ambiguity=ambiguity)

    def __init__(
            self, degrees, minutes=0, seconds=0,
            ambiguity=APRSPositionAmbiguity.NONE
    ):
        if degrees != int(degrees):
            # Decimal degrees given
            if minutes or seconds:
                raise ValueError(
                    "Cannot define minutes or seconds when "
                    "given decimal degrees"
                )
            idegrees = int(degrees)
            minutes = 60.0 * abs(degrees - idegrees)
            degrees = idegrees

        if (minutes < 0) or (minutes >= 60):
            raise ValueError("minutes must be in the range [0..59]")

        if minutes != int(minutes):
            # Decimal minutes given
            iminutes = int(minutes)
            if seconds:
                raise ValueError(
                    "Cannot define seconds when given decimal minutes"
                )
            seconds = 60.0 * (minutes - iminutes)
            minutes = iminutes

        if (seconds < 0) or (seconds >= 60):
            raise ValueError("seconds must be in the range [0..59]")

        self.degrees = degrees
        self.minutes = minutes
        self.seconds = seconds
        self.ambiguity = APRSPositionAmbiguity(ambiguity)

    def __repr__(self): # pragma: no cover
        return (
                '%s(degrees=%r, minutes=%r, seconds=%r, ambiguity=%r)' \
                % (
                    self.__class__.__name__,
                    self.degrees, self.minutes, self.seconds, self.ambiguity
                )
        )

    def __str__(self):
        """
        Return the APRS-formatted position string.
        """
        sign = self.NEG_SUFFIX if self.degrees < 0 else self.POS_SUFFIX
        degrees = abs(self.degrees)

        pos = self.STR_FORMAT % (
                degrees,
                self.minutes,
                self.seconds / 0.6,
                sign
        )

        if self.ambiguity.value > 0:
            pos = pos[0:-2] + ' ' + pos[-1:]
        if self.ambiguity.value > 1:
            pos = pos[0:-3] + ' ' + pos[-2:]
        if self.ambiguity.value > 2:
            pos = pos[0:-5] + ' ' + pos[-4:]
        if self.ambiguity.value > 3:
            pos = pos[0:-6] + ' ' + pos[-5:]

        return pos

    @property
    def decimalminutes(self):
        """
        Return the degrees and decimal minutes.
        """
        return self.minutes + (self.seconds / 60.0)

    @property
    def decimaldegrees(self):
        """
        Return the decimal degrees.
        """
        minutes = self.decimalminutes
        if self.degrees < 0:
            minutes = -minutes
        return self.degrees + (minutes / 60.0)


class APRSLatitude(APRSSexagesimal):
    POS_SUFFIX = "N"
    NEG_SUFFIX = "S"
    STR_FORMAT = "%02d%02d.%02d%s"
    LENGTH = 8


class APRSLongitude(APRSSexagesimal):
    POS_SUFFIX = "E"
    NEG_SUFFIX = "W"


class APRSUncompressedCoordinates(object):
    # Co-ordinate includes latitude, longitude, symbol table and
    # symbol
    LENGTH = APRSLatitude.LENGTH + APRSLongitude.LENGTH + 2

    @classmethod
    def decode(cls, coordinate):
        if len(coordinate) < cls.LENGTH:
            raise ValueError("Co-ordinate string is too short")

        # Strip extra data
        coordinate = coordinate[0:cls.LENGTH]

        # Decode and validate the symbol
        symbol = APRSSymbol(
                coordinate[APRSLatitude.LENGTH],
                coordinate[-1]
        )

        # Grab the latitude and longitude
        lat = APRSLatitude.decode(coordinate[0:APRSLatitude.LENGTH])
        lng = APRSLongitude.decode(coordinate[-APRSLongitude.LENGTH-1:-1])

        return cls(lat, lng, symbol)

    def __init__(self, lat, lng, symbol):
        self.lat = lat
        self.lng = lng
        self.symbol = symbol

    def __repr__(self): # pragma: no cover
        return (
                '%s(lat=%r, lng=%r, symbol=%r)' % \
                (
                        self.__class__.__name__,
                        self.lat,
                        self.lng,
                        self.symbol
                )
        )

    def __str__(self):
        return ''.join([
                str(self.lat),
                self.symbol.tableident,
                str(self.lng),
                self.symbol.symbol
        ])


class APRSCompressionTypeGPSFix(IntEnum):
    OLD     = 0b00000000
    CURRENT = 0b00100000


class APRSCompressionTypeNMEASrc(IntEnum):
    OTHER   = 0b00000000
    GLL     = 0b00001000
    GGA     = 0b00010000
    RMC     = 0b00011000


class APRSCompressionTypeOrigin(IntEnum):
    COMPRESSED  = 0b00000000
    TNC_BTEXT   = 0b00000001
    SOFTWARE    = 0b00000010
    TBD         = 0b00000011
    KPC3        = 0b00000100
    PICO        = 0b00000101
    OTHER       = 0b00000110
    DIGIPEATER  = 0b00000111


class APRSCompressionType(object):
    LENGTH = 1

    GPSFIX_MASK     = 0b00100000
    NMEASRC_MASK    = 0b00011000
    ORIGIN_MASK     = 0b00000111

    @classmethod
    def decode(cls, typechar):
        typebyte = ord(typechar) - BYTE_VALUE_OFFSET

        gpsfix = APRSCompressionTypeGPSFix(typebyte & cls.GPSFIX_MASK)
        nmeasrc = APRSCompressionTypeNMEASrc(typebyte & cls.NMEASRC_MASK)
        origin = APRSCompressionTypeOrigin(typebyte & cls.ORIGIN_MASK)

        return cls(gpsfix, nmeasrc, origin)


    def __init__(self, gpsfix, nmeasrc, origin):
        self.gpsfix = gpsfix
        self.nmeasrc = nmeasrc
        self.origin = origin

    @property
    def raw(self):
        return self.gpsfix.value | self.nmeasrc.value | self.origin.value

    def __repr__(self): # pragma: no cover
        return (
                '%s(gpsfix=%r, nmeasrc=%r, origin=%r)' \
                        % (
                            self.__class__.__name__,
                            self.gpsfix, self.nmeasrc, self.origin
                        )
        )

    def __str__(self):
        return chr(self.raw + BYTE_VALUE_OFFSET)


class APRSCompressedCoordinate(APRSSexagesimal):
    RADIX = 91
    LENGTH = 4

    @classmethod
    def decode(cls, coordinate):
        if len(coordinate) < cls.LENGTH:
            raise ValueError("Compressed co-ordinate too short")

        # Throw away extra data
        coordinate = coordinate[0:cls.LENGTH]

        # Convert the value back to decimal degrees
        value = ( \
                (decompress(coordinate) / cls.POSTSCALE) \
                - cls.OFFSET \
        ) / cls.PRESCALE

        return cls(value)

    def __str__(self):
        # Scale the decimal value
        return compress(int(abs(
            (
                (self.decimaldegrees * self.PRESCALE) + self.OFFSET
            ) * self.POSTSCALE
        )), self.LENGTH)


class APRSCompressedLatitude(APRSCompressedCoordinate):
    POSTSCALE = 380926
    PRESCALE = -1
    OFFSET = 90


class APRSCompressedLongitude(APRSCompressedCoordinate):
    POSTSCALE = 190463
    PRESCALE = 1
    OFFSET = 180


class APRSCompressedCourseSpeedRange(object):
    LENGTH = 2
    RADIX = APRSCompressedCoordinate.RADIX
    COURSE_SPEED_MAX = 89
    COURSE_SCALE = 4

    # this is in "knots".
    SPEED_UNITS = "knot"
    SPEED_RADIX = 1.08
    SPEED_OFFSET = -1

    # these values compute "miles".
    RANGE_UNITS = "mile"
    RANGE_HEADER = 90
    RANGE_SCALE = 2
    RANGE_RADIX = 1.08

    # these values compute "feet".
    ALTITUDE_UNITS = "foot"
    ALTITUDE_RADIX = 1.002

    @classmethod
    def decode(cls, csvalue, ctype):
        if len(csvalue) < cls.LENGTH:
            raise ValueError("Course/Speed value too short")

        # Discard extra data
        csvalue = csvalue[0:cls.LENGTH]

        if ctype.nmeasrc == APRSCompressionTypeNMEASrc.GGA:
            # This is an altitude value
            altitude_exp = decompress(csvalue)
            altitude = cls.ALTITUDE_RADIX ** altitude_exp
            speed = None
            course = None
            rng = None
        else:
            altitude = None
            csvalue = [
                    b - BYTE_VALUE_OFFSET
                    for b in bytes(csvalue, "us-ascii")
            ]

            if csvalue[0] == cls.RANGE_HEADER:
                # This is a range value
                rng = cls.RANGE_SCALE * (cls.RANGE_RADIX ** csvalue[1])
                speed = None
                course = None
            elif csvalue[0] <= cls.COURSE_SPEED_MAX:
                # This is a speed/course
                rng = None
                course = cls.COURSE_SCALE * csvalue[0]
                speed = (cls.SPEED_RADIX ** csvalue[1]) + cls.SPEED_OFFSET
            else:
                raise ValueError("Unknown Course/Speed/Range field: %r" \
                        % csvalue)

        return cls(course, speed, rng, altitude)

    def __init__(self, course=None, speed=None, rng=None, altitude=None):
        # Assert value units
        speed = convertvalue("speed", speed, self.SPEED_UNITS)
        rng = convertvalue("rng", rng, self.RANGE_UNITS)
        altitude = convertvalue("altitude", altitude, self.ALTITUDE_UNITS)

        if altitude is not None:
            for param in (course, speed, rng):
                if param is not None:
                    raise ValueError(
                        "Altitude cannot be specified with "
                        "range, course or speed"
                    )
        elif rng is not None:
            for param in (course, speed):
                if param is not None:
                    raise ValueError(
                        "Range cannot be specified with "
                        "altitude, course or speed"
                    )
        else:
            if (course is None) or (speed is None):
                raise ValueError(
                        "Course and speed must both be specified"
                )

        self.course = course
        self.speed = speed
        self.rng = rng
        self.altitude = altitude

    @property
    def speed_q(self):
        """
        Speed as a Pint quantity.
        """
        return Quantity(self.speed, self.SPEED_UNITS)

    @speed_q.setter
    def speed_q(self, value):
        """
        Set the speed from the given quantity.
        """
        self.speed = value.to(self.SPEED_UNITS).magnitude

    @property
    def rng_q(self):
        """
        Range as a Pint quantity.
        """
        return Quantity(self.rng, self.RANGE_UNITS)

    @rng_q.setter
    def rng_q(self, value):
        """
        Set the range from the given quantity.
        """
        self.rng = value.to(self.RANGE_UNITS).magnitude

    @property
    def altitude_q(self):
        """
        Altitude as a Pint quantity.
        """
        return Quantity(self.altitude, self.ALTITUDE_UNITS)

    @altitude_q.setter
    def altitude_q(self, value):
        """
        Set the altitude from the given quantity.
        """
        self.altitude = value.to(self.ALTITUDE_UNITS).magnitude

    def __repr__(self): # pragma: no cover
        return (
                '%s(course=%r, speed=%r, rng=%r, altitude=%r)' \
                % (
                    self.__class__.__name__,
                    self.course, self.speed, self.rng, self.altitude
                )
        )

    def __str__(self):
        if self.altitude is not None:
            return compress(
                    int(math.log(self.altitude, self.ALTITUDE_RADIX)),
                    self.LENGTH
            )
        elif self.rng is not None:
            # Return range
            bvalues = [
                    self.RANGE_HEADER,
                    int(math.log(
                        self.rng / self.RANGE_SCALE,
                        self.RANGE_RADIX
                    ))
            ]
        else:
            # Return speed/course
            bvalues = [
                    int(self.course / self.COURSE_SCALE),
                    int(math.log(
                        self.speed + 1,
                        self.SPEED_RADIX
                    ))
            ]

        return ''.join([chr(b + BYTE_VALUE_OFFSET) for b in bvalues])


class APRSCompressedCoordinates(object):
    LENGTH = APRSCompressedLatitude.LENGTH \
            + APRSCompressedLongitude.LENGTH \
            + APRSCompressionType.LENGTH \
            + APRSCompressedCourseSpeedRange.LENGTH \
            + 2 # symbol table ident + symbol code

    CST_FILL = " sT"

    @classmethod
    def decode(cls, coordinate):
        # Sanity check the length
        if len(coordinate) < cls.LENGTH:
            raise ValueError("Co-ordinate too short for compressed format")

        # Strip off anything extra
        coordinate = coordinate[0:cls.LENGTH]

        # Decode the symbol
        symbol = APRSSymbol(coordinate[0], coordinate[9])

        # Decode the latitude and longitude
        lat = APRSCompressedLatitude.decode(coordinate[1:])
        lng = APRSCompressedLongitude.decode(coordinate[5:])

        # Do we have a Course/Speed/Range or Type field?
        if coordinate[10] == cls.CST_FILL[0]:
            # No compression type or course/speed/range
            ctype = None
            csr = None
        else:
            # Decode the compression type byte
            ctype = APRSCompressionType.decode(coordinate[-1])
            csr = APRSCompressedCourseSpeedRange.decode(coordinate[-3:], ctype)

        return cls(lat, lng, symbol, ctype, csr)

    def __init__(self, lat, lng, symbol, ctype=None, csr=None):
        if (ctype is None) != (csr is None):
            raise ValueError("If either ctype or csr is given, give both")

        self.lat = lat
        self.lng = lng
        self.symbol = symbol
        self.ctype = ctype
        self.csr = csr

    def __repr__(self): # pragma: no cover
        return (
                '%s(lat=%r, lng=%r, symbol=%r, ctype=%r, csr=%r)' % \
                (
                        self.__class__.__name__,
                        self.lat,
                        self.lng,
                        self.symbol,
                        self.ctype,
                        self.csr
                )
        )

    def __str__(self):
        return ''.join([
            self.symbol.tableident,
            str(self.lat),
            str(self.lng),
            self.symbol.symbol,
            str(self.csr) if self.csr else self.CST_FILL,
            str(self.ctype) if self.ctype else ""
        ])


class APRSPositionFrame(APRSFrame):
    @classmethod
    def decode(cls, uiframe, payload, log):
        try:
            msgtype = APRSDataType(ord(payload[0]))
            log.debug("Received a %s frame", msgtype)
        except ValueError:
            raise ValueError('Not a recognised frame: %r' % payload)

        # If there's a timestamp, it'll be the first 7 characters
        if msgtype in (
                APRSDataType.POSITION_TS,
                APRSDataType.POSITION_TS_MSGCAP
        ):
            position_ts = decode_datetime(payload[1:8])
            payload = payload[8:]
        elif msgtype in (
                APRSDataType.POSITION,
                APRSDataType.POSITION_MSGCAP
        ):
            position_ts = None
            payload = payload[1:]
        else:
            raise ValueError('Not a position frame: %r' % payload)

        log.debug("Type: %s; timestamp: %r; payload: %r", \
                msgtype, position_ts, payload)

        # Is the position compressed?
        # Uncompressed position looks like this:
        #   0123456789012345678
        #   DDMM.MMsTDDDMM.MMsC
        #
        # whereas compressed looks like this:
        #   0123456789012
        #   TYYYYXXXXCcst
        #
        # D = degrees, M = minutes, s = sign, T = symbol table
        # C = symbol code, X = compressed longitude, Y = compressed latitude,
        # cs = course/speed, t = compression type
        #
        # The length needs to be at least the length of an uncompressed
        # position report (19 bytes) and must have decimal places in the
        # expected places (position 4 and 14).
        if (
            (len(payload) >= APRSUncompressedCoordinates.LENGTH) \
            and (payload[4] == ".") \
            and (payload[14] == ".")
        ):
            # This is possibly a position report
            position = APRSUncompressedCoordinates.decode(payload)
            message = payload[APRSUncompressedCoordinates.LENGTH:]
        else:
            # Possibly a compressed position report
            position = APRSCompressedCoordinates.decode(payload)
            message = payload[APRSCompressedCoordinates.LENGTH:]

        log.debug("Position: %r, Message: %r", position, message)

        return cls(
                destination=uiframe.header.destination,
                source=uiframe.header.source,
                position=position,
                position_ts=position_ts,
                message=message or None,
                messaging=(msgtype in (
                    APRSDataType.POSITION_MSGCAP,
                    APRSDataType.POSITION_TS_MSGCAP
                )),
                repeaters=uiframe.header.repeaters,
                pf=uiframe.pf,
                cr=uiframe.header.cr,
                src_cr=uiframe.header.src_cr
        )

    @classmethod
    def _encodepayload(
            cls, position, message=None, position_ts=None, messaging=True
        ):
        """
        Encode the message payload.
        """
        parts = []

        # Initial message type indicator
        if position_ts is None:
            if messaging:
                parts.append(chr(APRSDataType.POSITION_MSGCAP.value))
            else:
                parts.append(chr(APRSDataType.POSITION.value))
        else:
            if messaging:
                parts.append(chr(APRSDataType.POSITION_TS_MSGCAP.value))
            else:
                parts.append(chr(APRSDataType.POSITION_TS.value))

            # Timestamp
            parts.append(str(position_ts))

        # Position
        parts.append(str(position))

        # Message (Comment)
        if message is not None:
            parts.append(str(message))

        return ''.join(parts)

    def __init__(self, destination, source, position, position_ts=None,
            message=None, messaging=True, repeaters=None,
            pf=False, cr=True, src_cr=None):
        self._position = position
        self._position_ts = position_ts
        self._messaging = messaging
        self._message = message

        payload = self._encodepayload(position, message, position_ts, messaging)

        super(APRSPositionFrame, self).__init__(
                destination=destination,
                source=source,
                payload=payload.encode('US-ASCII'),
                repeaters=repeaters, pf=pf, cr=cr, src_cr=src_cr)

    @property
    def position(self):
        return self._position

    @property
    def position_ts(self):
        return self._position_ts

    @property
    def message(self):
        return self._message

    @property
    def has_messaging(self):
        return self._messaging


APRSFrame.DATA_TYPE_HANDLERS[APRSDataType.POSITION] = \
        APRSPositionFrame
APRSFrame.DATA_TYPE_HANDLERS[APRSDataType.POSITION_MSGCAP] = \
        APRSPositionFrame
APRSFrame.DATA_TYPE_HANDLERS[APRSDataType.POSITION_TS] = \
        APRSPositionFrame
APRSFrame.DATA_TYPE_HANDLERS[APRSDataType.POSITION_TS_MSGCAP] = \
        APRSPositionFrame



#!/usr/bin/env python3

"""
Base KISS interface unit tests.
"""

from aioax25.kiss import BaseKISSDevice, KISSDeviceState, KISSCommand, KISSPort
from ..loop import DummyLoop
from asyncio import BaseEventLoop

from ..nosecompat import eq_, assert_is, assert_greater, assert_less


class DummyKISSDevice(BaseKISSDevice):
    def __init__(self, **kwargs):
        super(DummyKISSDevice, self).__init__(**kwargs)

        self.transmitted = bytearray()
        self.open_calls = 0
        self.close_calls = 0

    def _open(self):
        self.open_calls += 1

    def _close(self):
        self.close_calls += 1

    def _send_raw_data(self, data):
        self.transmitted += data


def test_constructor_own_loop():
    """
    Test constructor uses its own IOLoop if not given one
    """
    kissdev = DummyKISSDevice(loop=None)
    assert isinstance(kissdev._loop, BaseEventLoop)


def test_open():
    """
    Test an open call passes to subclass _open
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop)

    eq_(kissdev.open_calls, 0)
    kissdev.open()

    eq_(kissdev.open_calls, 1)

def test_close():
    """
    Test a close call passes to _close
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=False
    )

    # Force the port open
    kissdev._state = KISSDeviceState.OPEN

    eq_(kissdev.close_calls, 0)

    # Now try closing the port
    kissdev.close()
    eq_(kissdev.close_calls, 1)

def test_close_reset():
    """
    Test a close call with reset_on_close sends the "return from KISS" frame
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=True
    )

    # Force the port open
    kissdev._state = KISSDeviceState.OPEN

    # Now try closing the port
    kissdev.close()

    # Should be in the closing state
    eq_(kissdev._state, KISSDeviceState.CLOSING)

    # A "return from KISS" frame should be in the transmit buffer
    eq_(bytes(kissdev._tx_buffer), b'\xc0\xff\xc0')

    # A call to _send_data should be pending
    (_, func) = loop.calls.pop()
    eq_(func, kissdev._send_data)

def test_receive():
    """
    Test that a call to _receive stashes the data then schedules _receive_frame.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=True
    )
    kissdev._receive(b'test incoming data')

    # Data should be waiting
    eq_(bytes(kissdev._rx_buffer), b'test incoming data')

    # A call to _receive_frame should be pending
    (_, func) = loop.calls.pop()
    eq_(func, kissdev._receive_frame)

def test_receive_frame_garbage():
    """
    Test _receive_frame discards all data when no FEND byte found.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=True
    )
    kissdev._rx_buffer += b'this should be discarded'
    kissdev._receive_frame()

    # We should just have the data including and following the FEND
    eq_(bytes(kissdev._rx_buffer), b'')

    # As there's no complete frames, no calls should be scheduled
    eq_(len(loop.calls), 0)

def test_receive_frame_garbage_start():
    """
    Test _receive_frame discards everything up to the first FEND byte.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=True
    )
    kissdev._rx_buffer += b'this should be discarded\xc0this should be kept'
    kissdev._receive_frame()

    # We should just have the data including and following the FEND
    eq_(bytes(kissdev._rx_buffer), b'\xc0this should be kept')

    # As there's no complete frames, no calls should be scheduled
    eq_(len(loop.calls), 0)

def test_receive_frame_empty():
    """
    Test _receive_frame discards empty frames.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=True
    )
    kissdev._rx_buffer += b'\xc0\xc0'
    kissdev._receive_frame()

    # We should just have the last FEND
    eq_(bytes(kissdev._rx_buffer), b'\xc0')

    # It should leave the last FEND there and wait for more data.
    eq_(len(loop.calls), 0)

def test_receive_frame_single():
    """
    Test _receive_frame hands a single frame to _dispatch_rx_frame.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=True
    )
    kissdev._rx_buffer += b'\xc0\x00a single KISS frame\xc0'
    kissdev._receive_frame()

    # We should just have the last FEND
    eq_(bytes(kissdev._rx_buffer), b'\xc0')

    # We should have one call to _dispatch_rx_frame
    eq_(len(loop.calls), 1)
    (_, func, frame) = loop.calls.pop(0)
    eq_(func, kissdev._dispatch_rx_frame)
    assert isinstance(frame, KISSCommand)
    eq_(frame.port, 0)
    eq_(frame.cmd, 0)
    eq_(frame.payload, b'a single KISS frame')

def test_receive_frame_more():
    """
    Test _receive_frame calls itself when more data left.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(
            loop=loop, reset_on_close=True
    )
    kissdev._rx_buffer += b'\xc0\x00a single KISS frame\xc0more data'
    kissdev._receive_frame()

    # We should just have the left-over bit including the last FEND
    eq_(bytes(kissdev._rx_buffer), b'\xc0more data')

    # This should have generated two calls:
    eq_(len(loop.calls), 2)

    # We should have one call to _dispatch_rx_frame
    (_, func, frame) = loop.calls.pop(0)
    eq_(func, kissdev._dispatch_rx_frame)
    assert isinstance(frame, KISSCommand)
    eq_(frame.port, 0)
    eq_(frame.cmd, 0)
    eq_(frame.payload, b'a single KISS frame')

    # We should have another to _receive_frame itself.
    (_, func) = loop.calls.pop(0)
    eq_(func, kissdev._receive_frame)

def test_dispatch_rx_invalid_port():
    """
    Test that _dispatch_rx_port to an undefined port drops the frame.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop)
    kissdev._dispatch_rx_frame(
            KISSCommand(cmd=10, port=14, payload=b'this should be dropped')
    )

def test_dispatch_rx_exception():
    """
    Test that _dispatch_rx_port drops frame on exception.
    """
    class DummyPort(object):
        def _receive_frame(self, frame):
            raise IOError('Whoopsie')

    port = DummyPort()
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop)
    kissdev._port[14] = port

    # Deliver the frame
    frame = KISSCommand(cmd=10, port=14, payload=b'this should be dropped')
    kissdev._dispatch_rx_frame(frame)

def test_dispatch_rx_valid_port():
    """
    Test that _dispatch_rx_port to a known port delivers to that port.
    """
    class DummyPort(object):
        def __init__(self):
            self.frames = []

        def _receive_frame(self, frame):
            self.frames.append(frame)

    port = DummyPort()
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop)
    kissdev._port[14] = port

    # Deliver the frame
    frame = KISSCommand(cmd=10, port=14, payload=b'this should be delivered')
    kissdev._dispatch_rx_frame(frame)

    # Our port should have the frame
    eq_(len(port.frames), 1)
    assert_is(port.frames[0], frame)

def test_send_data():
    """
    Test that _send_data sends whatever data is buffered up to the block size.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop)
    kissdev._tx_buffer += b'test output data'

    # Send the data out.
    kissdev._send_data()

    # We should now see this was "sent" and now in 'transmitted'
    eq_(bytes(kissdev.transmitted), b'test output data')

    # That should be the lot
    eq_(len(loop.calls), 0)

def test_send_data_block_size():
    """
    Test that _send_data re-schedules itself when buffer exceeds block size
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop, send_block_size=4, send_block_delay=1)
    kissdev._tx_buffer += b'test output data'

    # Send the data out.
    kissdev._send_data()

    # We should now see the first block was "sent" and now in 'transmitted'
    eq_(bytes(kissdev.transmitted), b'test')

    # The rest should be waiting
    eq_(bytes(kissdev._tx_buffer), b' output data')

    # There should be a pending call to send more:
    eq_(len(loop.calls), 1)
    (calltime, callfunc) = loop.calls.pop(0)

    # It'll be roughly in a second's time calling the same function
    assert_greater(calltime - loop.time(), 0.990)
    assert_less(calltime - loop.time(), 1.0)
    eq_(callfunc, kissdev._send_data)

def test_send_data_close_after_send():
    """
    Test that _send_data when closing the device, closes after last send
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop)
    kissdev._tx_buffer += b'test output data'

    # Force state
    kissdev._state = KISSDeviceState.CLOSING

    # No close call made yet
    eq_(kissdev.close_calls, 0)

    # Send the data out.
    kissdev._send_data()

    # We should now see the first block was "sent" and now in 'transmitted'
    eq_(bytes(kissdev.transmitted), b'test output data')

    # The device should now be closed.
    eq_(kissdev.close_calls, 1)

def test_send_data_block_size():
    """
    Test that _send_data waits until all data sent before closing.
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop, send_block_size=4, send_block_delay=1)
    kissdev._tx_buffer += b'test output data'

    # Force state
    kissdev._state = KISSDeviceState.CLOSING

    # Send the data out.
    kissdev._send_data()

    # We should now see the first block was "sent" and now in 'transmitted'
    eq_(bytes(kissdev.transmitted), b'test')

    # The rest should be waiting
    eq_(bytes(kissdev._tx_buffer), b' output data')

    # There should be a pending call to send more:
    eq_(len(loop.calls), 1)
    (calltime, callfunc) = loop.calls.pop(0)

    # It'll be roughly in a second's time calling the same function
    assert_greater(calltime - loop.time(), 0.990)
    assert_less(calltime - loop.time(), 1.0)
    eq_(callfunc, kissdev._send_data)

    # No close call made yet
    eq_(kissdev.close_calls, 0)

def test_init_kiss():
    """
    Test _init_kiss sets up the commands to be sent to initialise KISS
    """
    loop = DummyLoop()
    kissdev = DummyKISSDevice(loop=loop)

    # Force state
    kissdev._state = KISSDeviceState.OPENING

    # Initialise the KISS device
    kissdev._init_kiss()

    # We should see a copy of the KISS commands, minus the first
    eq_(kissdev._kiss_rem_commands, kissdev._kiss_commands[1:])

    # We should see the first initialisation commands
    eq_(bytes(kissdev.transmitted), b'INT KISS\r')

def test_getitem():
    """
    Test __getitem__ returns a port instance.
    """
    kissdev = DummyKISSDevice(loop=DummyLoop())
    port = kissdev[7]
    assert isinstance(port, KISSPort)
    assert_is(kissdev._port[7], port)

def test__send_kiss_cmd():
    kissdev = DummyKISSDevice(loop=DummyLoop())
    kissdev._kiss_rem_commands = []

    kissdev._send_kiss_cmd()
    eq_(KISSDeviceState.OPEN, kissdev._state)
    eq_(bytearray(), kissdev._rx_buffer)




#!/usr/bin/env python3

"""
APRS compression algorithm.
"""

BYTE_VALUE_OFFSET = 33
BYTE_VALUE_RADIX = 91


def compress(value, length):
    # Initialise our byte values
    bvalue = [0] * length

    # Figure out the bytes
    for pos in range(length):
        (div, rem) = divmod(
                value,
                BYTE_VALUE_RADIX ** (length - pos - 1)
        )
        bvalue[pos] += int(div)
        value = rem

    # Encode them into ASCII
    return ''.join([chr(b + BYTE_VALUE_OFFSET) for b in bvalue])


def decompress(value):
    length = len(value)
    return sum([
            (
                (b - BYTE_VALUE_OFFSET)
                * (BYTE_VALUE_RADIX ** (length - i - 1))
            )
            for (i, b) in enumerate(bytes(value, "us-ascii"))
    ])



#!/usr/bin/env python3

"""
TCP KISS interface unit tests.
"""

# Most of the functionality here is common to SerialKISSDevice, so this
# really just tests that we pass the right commands to the IOLoop when
# establishing a connection.

from aioax25 import kiss
import logging
from ..asynctest import asynctest
from asyncio import get_event_loop, sleep


@asynctest
async def test_open_connection():
    # This will receive the arguments passed to create_connection
    connection_args = {}

    loop = get_event_loop()

    # Stub the create_connection method
    orig_create_connection = loop.create_connection
    async def _create_connection(proto_factory, **kwargs):
        # proto_factory should give us a KISSProtocol object
        protocol = proto_factory()
        assert isinstance(protocol, kiss.KISSProtocol)

        connection_args.update(kwargs)
    loop.create_connection = _create_connection

    try:
        device = kiss.TCPKISSDevice(
                host='localhost', port=5432,
                loop=loop, log=logging.getLogger(__name__)
        )

        await device._open_connection()

        # Expect a connection attempt to have been made
        assert connection_args == dict(
            host='localhost', port=5432,
            ssl=None, family=0, proto=0, flags=0,
            sock=None, local_addr=None, server_hostname=None
        )
    finally:
        # Restore mock
        loop.create_connection = orig_create_connection



#!/usr/bin/env python3

"""
KISS protocol unit tests.
"""

from aioax25 import kiss
import logging
import time


class DummyTransport(object):
    def __init__(self):
        self.closed = False

    def close(self):
        assert self.closed is False, 'Already closed'
        self.closed = True


class TestError(Exception):
    pass


def test_protocol_connection_made(logger):
    """
    Test connection_made calls the _on_connect call-back.
    """
    on_connect_calls = []
    transport = DummyTransport()


    def on_connect(*args, **kwargs):
        on_connect_calls.append((args, kwargs))

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSProtocol(on_connect, on_receive, on_close, logger)
    protocol.connection_made(transport)

    assert on_connect_calls == [((transport,), {})]
    assert transport.closed is False
    assert logger.logrecords == []
    assert logger.children == {}


def test_protocol_connection_made_err(logger):
    """
    Test connection_made handles errors.
    """
    on_connect_calls = []
    transport = DummyTransport()

    def on_connect(*args, **kwargs):
        on_connect_calls.append((args, kwargs))
        raise TestError()

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSProtocol(on_connect, on_receive, on_close, logger)
    protocol.connection_made(transport)

    assert logger.children == {}
    assert len(logger.logrecords) > 0
    assert logger.logrecords[1:] == []

    log = logger.logrecords[0]
    assert log.pop('ex_type', None) == TestError
    log.pop('ex_val', None)
    log.pop('ex_tb', None)

    assert log == dict(
        method='exception',
        args=('Failed to handle connection establishment',),
        kwargs={}
    )


def test_protocol_data_received(logger):
    """
    Test data_received calls the _on_connect call-back.
    """
    on_receive_calls = []
    received_data = b'Dummy received data'

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        on_receive_calls.append((args, kwargs))

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSProtocol(on_connect, on_receive, on_close,
            logging.getLogger('%s.data_received' % __name__))

    protocol.data_received(received_data)

    assert on_receive_calls == [((received_data,), {})]
    assert logger.logrecords == []
    assert logger.children == {}


def test_protocol_data_received_err(logger):
    """
    Test data_received handles errors.
    """
    on_receive_calls = []
    received_data = b'Dummy received data'

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        on_receive_calls.append((args, kwargs))
        raise TestError()

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSProtocol(on_connect, on_receive, on_close, logger)
    protocol.data_received(received_data)

    assert on_receive_calls == [((received_data,), {})]
    assert logger.children == {}
    assert len(logger.logrecords) > 0
    assert logger.logrecords[1:] == []

    log = logger.logrecords[0]
    assert log.pop('ex_type', None) == TestError
    log.pop('ex_val', None)
    log.pop('ex_tb', None)

    assert log == dict(
        method='exception',
        args=('Failed to handle incoming data',),
        kwargs={}
    )


def test_protocol_connection_lost(logger):
    """
    Test connection_lost calls the _on_connect call-back.
    """
    on_close_calls = []
    loss_err = TestError()

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        on_close_calls.append((args, kwargs))

    protocol = kiss.KISSProtocol(on_connect, on_receive, on_close, logger)
    protocol.connection_lost(loss_err)

    assert on_close_calls == [((loss_err,), {})]
    assert logger.logrecords == []
    assert logger.children == {}


def test_protocol_connection_lost_err(logger):
    """
    Test connection_lost handles errors.
    """
    class TestConnectionLossError(Exception):
        pass

    on_close_calls = []
    loss_err = TestConnectionLossError()

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        on_close_calls.append((args, kwargs))
        raise TestError()

    protocol = kiss.KISSProtocol(on_connect, on_receive, on_close, logger)
    protocol.connection_lost(loss_err)

    assert on_close_calls == [((loss_err,), {})]
    assert logger.children == {}
    assert len(logger.logrecords) > 0
    assert logger.logrecords[1:] == []

    log = logger.logrecords[0]
    assert log.pop('ex_type', None) == TestError
    log.pop('ex_val', None)
    log.pop('ex_tb', None)

    assert log == dict(
        method='exception',
        args=('Failed to handle connection loss',),
        kwargs={}
    )


def test_subproc_protocol_connection_made(logger):
    """
    Test connection_made calls the _on_connect call-back.
    """
    on_connect_calls = []
    transport = DummyTransport()


    def on_connect(*args, **kwargs):
        on_connect_calls.append((args, kwargs))

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSSubprocessProtocol(on_connect, on_receive, on_close, logger)
    protocol.connection_made(transport)

    assert on_connect_calls == [((transport,), {})]
    assert transport.closed is False
    assert logger.logrecords == []
    assert logger.children == {}


def test_subproc_protocol_connection_made_err(logger):
    """
    Test connection_made handles errors.
    """
    on_connect_calls = []
    transport = DummyTransport()

    def on_connect(*args, **kwargs):
        on_connect_calls.append((args, kwargs))
        raise TestError()

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSSubprocessProtocol(on_connect, on_receive, on_close, logger)
    protocol.connection_made(transport)

    assert logger.children == {}
    assert len(logger.logrecords) > 0
    assert logger.logrecords[1:] == []

    log = logger.logrecords[0]
    assert log.pop('ex_type', None) == TestError
    log.pop('ex_val', None)
    log.pop('ex_tb', None)

    assert log == dict(
        method='exception',
        args=('Failed to handle connection establishment',),
        kwargs={}
    )


def test_subproc_protocol_pipe_data_received_stdout(logger):
    """
    Test pipe_data_received calls the _on_connect call-back for stdout.
    """
    on_receive_calls = []
    received_data = b'Dummy received data'

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        on_receive_calls.append((args, kwargs))

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSSubprocessProtocol(
            on_connect, on_receive, on_close, logger
    )
    protocol.pipe_data_received(1, received_data)

    assert on_receive_calls == [((received_data,), {})]
    assert logger.logrecords == []
    assert logger.children == {}


def test_subproc_protocol_pipe_data_received_stderr(logger):
    """
    Test pipe_data_received ignores data received on stderr.
    """
    on_receive_calls = []
    received_data = b'Dummy received data'

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSSubprocessProtocol(
            on_connect, on_receive, on_close, logger
    )
    protocol.pipe_data_received(2, received_data)

    assert on_receive_calls == []
    assert logger.children == {}
    assert logger.logrecords == [
            dict(
                method='debug',
                args=(
                    'Data received on fd=%d: %r',
                    2, received_data
                ),
                kwargs={},
                ex_type=None, ex_val=None, ex_tb=None
            )
    ]


def test_subproc_protocol_pipe_data_received_err(logger):
    """
    Test pipe_data_received handles errors.
    """
    on_receive_calls = []
    received_data = b'Dummy received data'

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        on_receive_calls.append((args, kwargs))
        raise TestError()

    def on_close(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    protocol = kiss.KISSSubprocessProtocol(on_connect, on_receive, on_close, logger)
    protocol.pipe_data_received(1, received_data)

    assert on_receive_calls == [((received_data,), {})]
    assert logger.children == {}
    assert len(logger.logrecords) > 0
    assert logger.logrecords[1:] == []

    log = logger.logrecords[0]
    assert log.pop('ex_type', None) == TestError
    log.pop('ex_val', None)
    log.pop('ex_tb', None)

    assert log == dict(
        method='exception',
        args=(
            'Failed to handle incoming data %r on fd=%d',
            received_data, 1
        ),
        kwargs={}
    )


def test_subproc_protocol_process_exited(logger):
    """
    Test process_exited calls the _on_connect call-back.
    """
    on_close_calls = []

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        on_close_calls.append((args, kwargs))

    protocol = kiss.KISSSubprocessProtocol(
            on_connect, on_receive, on_close, logger
    )

    protocol.process_exited()

    assert on_close_calls == [((None,), {})]
    assert logger.logrecords == []
    assert logger.children == {}


def test_subproc_protocol_process_exited_err(logger):
    """
    Test process_exited handles errors.
    """
    class TestConnectionLossError(Exception):
        pass

    on_close_calls = []
    loss_err = TestConnectionLossError()

    def on_connect(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_receive(*args, **kwargs):
        assert False, \
                'Should not have been called (called with %r, %r)' \
                % (args, kwargs)

    def on_close(*args, **kwargs):
        on_close_calls.append((args, kwargs))
        raise TestError()

    protocol = kiss.KISSSubprocessProtocol(on_connect, on_receive, on_close, logger)
    protocol.process_exited()

    assert on_close_calls == [((None,), {})]
    assert logger.children == {}
    assert len(logger.logrecords) > 0
    assert logger.logrecords[1:] == []

    log = logger.logrecords[0]
    assert log.pop('ex_type', None) == TestError
    log.pop('ex_val', None)
    log.pop('ex_tb', None)

    assert log == dict(
        method='exception',
        args=('Failed to handle process exit',),
        kwargs={}
    )



#!/usr/bin/env python3

"""
APRS date and time formats.
"""

from datetime import time


class APRSTimestamp(object):
    """
    Base abstract class for APRS timestamps.
    """
    pass


class DHMBaseTimestamp(APRSTimestamp):
    """
    Day/Hour/Minute timestamp (base class)
    """
    TS_LENGTH = 7

    def __init__(self, day, hour, minute):
        self.day = day
        self.hour = hour
        self.minute = minute

    def __str__(self):
        return '%02d%02d%02d%s' % (
                self.day, self.hour, self.minute,
                self.TS_SUFFIX
        )


class DHMUTCTimestamp(DHMBaseTimestamp):
    """
    Day/Hour/Minute timestamp in UTC.
    """
    TS_SUFFIX = "z"


class DHMLocalTimestamp(DHMBaseTimestamp):
    """
    Day/Hour/Minute timestamp in local time.
    """
    TS_SUFFIX = "/"


class HMSTimestamp(time, APRSTimestamp):
    """
    Hour/Minute/Second timestamp in UTC.
    """
    TS_LENGTH = 7
    TS_SUFFIX = "h"

    def __str__(self):
        return '%02d%02d%02d%s' % (
                self.hour, self.minute, self.second,
                self.TS_SUFFIX
        )


class MDHMTimestamp(APRSTimestamp):
    """
    Month/Day/Hour/Minute timestamp in UTC.
    """
    TS_LENGTH = 8

    def __init__(self, month, day, hour, minute):
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute

    def __str__(self):
        return '%02d%02d%02d%02d' % (
                self.month, self.day, self.hour, self.minute
        )


def decode(dtstr):
    """
    Decode a APRS date/time date code.  Since not full information is given
    in the timestamp, and date/time values are a nightmare to handle
    (especially for local timestamps), we just reproduce enough to round-trip
    the values in APRS.
    """
    if len(dtstr) < DHMBaseTimestamp.TS_LENGTH:
        # Not a valid date/time format
        raise ValueError("Timestamp string too short")

    if dtstr[6] in (DHMLocalTimestamp.TS_SUFFIX, DHMUTCTimestamp.TS_SUFFIX):
        # Day/Hours/Minutes in UTC or local-time
        # Format is:
        #   DDHHMM{z|/}
        day = int(dtstr[0:2])
        hour = int(dtstr[2:4])
        minute = int(dtstr[4:6])

        if dtstr[6] == DHMUTCTimestamp.TS_SUFFIX:
            return DHMUTCTimestamp(day, hour, minute)
        else:
            return DHMLocalTimestamp(day, hour, minute)
    elif dtstr[6] == HMSTimestamp.TS_SUFFIX:
        # Hours/Minutes/Seconds in UTC
        # Format is:
        #   HHMMSSh
        hour = int(dtstr[0:2])
        minute = int(dtstr[2:4])
        second = int(dtstr[4:6])

        return HMSTimestamp(hour, minute, second)
    elif len(dtstr) >= MDHMTimestamp.TS_LENGTH:
        # Month/Day/Hours/Minutes in UTC
        # Format is:
        #   MMDDHHMM
        month = int(dtstr[0:2])
        day = int(dtstr[2:4])
        hour = int(dtstr[4:6])
        minute = int(dtstr[6:8])

        return MDHMTimestamp(month, day, hour, minute)

    raise ValueError("Timestamp format not recognised")



#!/usr/bin/env python3

"""
AX.25 Interface handler
"""

import logging
import asyncio
import random
from functools import partial
import time

from .router import Router


class AX25Interface(Router):
    """
    The AX25Interface class represents a logical AX.25 interface.
    The interface handles basic queueing and routing of message traffic.

    Outgoing messages are queued and sent when there is a break of greater
    than the cts_delay (10ms) + a randomisation factor (cts_rand).
    Messages may be cancelled prior to transmission.
    """

    def __init__(self, kissport, cts_delay=0.01,
            cts_rand=0.01, log=None, loop=None):
        # Initialise the superclass
        super(AX25Interface, self).__init__()

        if log is None:
            log = logging.getLogger(self.__class__.__module__)

        if loop is None:
            loop = asyncio.get_event_loop()

        self._log = log
        self._loop = loop
        self._port = kissport

        # Message queue
        self._tx_queue = []
        self._tx_pending = None

        # Clear-to-send delay and randomisation factor
        self._cts_delay = cts_delay
        self._cts_rand = cts_rand

        # Clear-to-send expiry
        self._cts_expiry = loop.time() \
                + cts_delay \
                + (random.random() * cts_rand)

        # Bind to the KISS port to receive raw messages.
        kissport.received.connect(self._on_receive)

    def transmit(self, frame, callback=None):
        """
        Enqueue a message for transmission.  Optionally give a call-back
        function to receive notification of transmission.
        """
        self._log.debug('Adding to queue: %s', frame)
        self._tx_queue.append((frame, callback))
        if not self._tx_pending:
            self._schedule_tx()

    def cancel_transmit(self, frame):
        """
        Cancel the transmission of a frame.
        """
        self._log.debug('Removing from queue: %s', frame)
        self._tx_queue = list(filter(
            lambda item : item[0] is not frame,
            self._tx_queue
        ))

    def _reset_cts(self):
        """
        Reset the clear-to-send timer.
        """
        cts_expiry = self._loop.time() \
                + self._cts_delay + (random.random() * self._cts_rand)

        # Ensure CTS expiry never goes backwards!
        while cts_expiry < self._cts_expiry:
            cts_expiry += (random.random() * self._cts_rand)
        self._cts_expiry = cts_expiry

        self._log.debug('Clear-to-send expiry at %s', self._cts_expiry)
        if self._tx_pending:
            # We were waiting for a clear-to-send, so re-schedule.
            self._schedule_tx()

    def _on_receive(self, frame):
        """
        Handle an incoming message.
        """
        self._reset_cts()
        super(AX25Interface, self)._on_receive(frame)

    def _schedule_tx(self):
        """
        Schedule the transmit timer to take place after the CTS expiry.
        """
        if self._tx_pending:
            self._tx_pending.cancel()

        delay = self._cts_expiry - self._loop.time()
        if delay > 0:
            self._log.debug('Scheduling next transmission in %s', delay)
            self._tx_pending = self._loop.call_later(delay, self._tx_next)
        else:
            self._log.debug('Scheduling next transmission ASAP')
            self._tx_pending = self._loop.call_soon(self._tx_next)

    def _tx_next(self):
        """
        Transmit the next message.
        """
        self._tx_pending = None

        try:
            (frame, callback) = self._tx_queue.pop(0)
        except IndexError:
            self._log.debug('No traffic to transmit')
            return

        try:
            if (frame.deadline is not None) and \
                    (frame.deadline < time.time()):
                self._log.info('Dropping expired frame: %s', frame)
                self._schedule_tx()
                return
        except AttributeError: # pragma: no cover
            # Technically, all objects that pass through here should be
            # AX25Frame sub-classes, so this branch should not get executed.
            # If it does, we just pretend there is no deadline.
            pass

        try:
            self._log.debug('Transmitting %s', frame)
            self._port.send(frame)
            if callback:
                self._log.debug('Notifying sender of %s', frame)
                self._loop.call_soon(partial(
                    callback,
                    interface=self,
                    frame=frame
                ))
        except:
            self._log.exception('Failed to transmit %s', frame)

        self._reset_cts()
        if len(self._tx_queue) > 0:
            self._schedule_tx()



#!/usr/bin/env python3

"""
AX.25 Interface handler
"""

from functools import partial
from .signal import Signal
import re

from .frame import AX25Frame


class Router(object):
    """
    The Router routes incoming messages to receivers.  It is a mix-in class
    used by AX25Interface.  The base class is assumed to have IOLoop and
    logger instances
    """

    def __init__(self):
        # Receivers
        self._receiver_str = {}
        self._receiver_re = {}

        # Received message call-back.  This is triggered whenever a message
        # comes in regardless of destination call-sign.
        self.received_msg = Signal()

    def _get_destination(self, frame):
        """
        Retrieve the destination of the given frame and return it.
        """
        return frame.header.destination

    def bind(self, callback, callsign, ssid=0, regex=False):
        """
        Bind a receiver to the given call-sign and optional SSID.  The callsign
        argument is expected to be a string, but may also be a regular
        expression pattern which is then matched against the callsign (but not
        the SSID!).

        ssid may be set to None, which means all SSIDs.
        """
        if not isinstance(callsign, str):
            raise TypeError('callsign must be a string (use '
                            'regex=True for regex)')
        if regex:
            (_, call_receivers) = self._receiver_re.setdefault(
                    callsign,
                    (re.compile(callsign), {})
            )
        else:
            call_receivers = self._receiver_str.setdefault(
                    callsign,
                    {}
            )

        call_receivers.setdefault(ssid, []).append(callback)
            
    def unbind(self, callback, callsign, ssid=0, regex=False):
        """
        Unbind a receiver from the given callsign/SSID combo.
        """
        try:
            if regex:
                receivers = self._receiver_re
                (_, call_receivers) = receivers[callsign]
            else:
                receivers = self._receiver_str
                call_receivers = receivers[callsign]

            ssid_receivers = call_receivers[ssid]
        except KeyError:
            return

        try:
            ssid_receivers.remove(callback)
        except ValueError:
            return

        if len(ssid_receivers) == 0:
            call_receivers.pop(ssid)

        if len(call_receivers) == 0:
            receivers.pop(callsign)

    def _on_receive(self, frame):
        """
        Handle an incoming message.
        """
        # Decode from raw bytes
        if not isinstance(frame, AX25Frame):
            frame = AX25Frame.decode(frame)
        self._log.debug('Handling incoming frame %s', frame)

        # Pass the message to those who elected to receive all traffic
        self._loop.call_soon(partial(
            self.received_msg.emit,
            interface=self, frame=frame))

        destination = self._get_destination(frame)
        callsign = destination.callsign
        ssid = destination.ssid

        # Dispatch the incoming message notification to string match receivers
        calls = []
        try:
            callsign_receivers = self._receiver_str[callsign]
            calls.extend([
                partial(receiver, interface=self, frame=frame)
                for receiver in
                callsign_receivers.get(None, []) \
                            + callsign_receivers.get(ssid, [])
            ])
        except KeyError:
            pass

        # Compare the incoming frame destination to our regex receivers
        for (pattern, pat_receivers) in self._receiver_re.values():
            match = pattern.search(callsign)
            if not match:
                continue

            calls.extend([
                partial(receiver, interface=self, frame=frame, match=match)
                for receiver in
                pat_receivers.get(None, []) \
                        + pat_receivers.get(ssid, [])
            ])

        # Dispatch the received message
        self._log.debug('Dispatching frame to %d receivers', len(calls))
        for receiver in calls:
            self._loop.call_soon(receiver)



#!/usr/bin/env python3

import logging

from aioax25.frame import AX25UnnumberedInformationFrame
from aioax25.unit import Quantity
from aioax25.aprs.datetime import MDHMTimestamp
from aioax25.aprs.position import APRSSexagesimal, \
        APRSLatitude, APRSLongitude, \
        APRSCompressedLatitude, APRSCompressedLongitude, \
        APRSPositionFrame, APRSPositionAmbiguity, \
        APRSUncompressedCoordinates, \
        APRSCompressedCourseSpeedRange, \
        APRSCompressedCoordinates, \
        APRSCompressionType, \
        APRSCompressionTypeGPSFix, \
        APRSCompressionTypeNMEASrc, \
        APRSCompressionTypeOrigin
from aioax25.aprs.symbol import APRSSymbol

"""
Position handling tests.
"""

def test_sexagesimal_convert_dd():
    """
    Test we can convert decimal degrees to sexagesimal format.
    """
    v = APRSSexagesimal(-27.437244)
    assert (v.degrees, v.minutes) == (-27, 26)
    assert abs(v.seconds - 14.0777) < 0.001

def test_sexagesimal_convert_dm():
    """
    Test we can convert decimal minutes to sexagesimal format.
    """
    v = APRSSexagesimal(-27, 26.234628333)
    assert (v.degrees, v.minutes) == (-27, 26)
    assert abs(v.seconds - 14.0777) < 0.001

def test_sexagesimal_convert_to_dm():
    """
    Test we can convert sexagesimal format to decimal minutes.
    """
    v = APRSSexagesimal(-27, 26, 14.0777)
    assert abs(v.decimalminutes - 26.2346) < 0.001

def test_sexagesimal_convert_to_dd():
    """
    Test we can convert sexagesimal format to decimal degrees.
    """
    v = APRSSexagesimal(-27, 26, 14.0777)
    assert abs(v.decimaldegrees - -27.437244) < 0.00001

def test_sexagesimal_convert_to_dm_pos():
    """
    Test we can convert sexagesimal format to decimal minutes. (Positive)
    """
    v = APRSSexagesimal(27, 26, 14.0777)
    assert abs(v.decimalminutes - 26.2346) < 0.001

def test_sexagesimal_convert_to_dd_pos():
    """
    Test we can convert sexagesimal format to decimal degrees. (Positive)
    """
    v = APRSSexagesimal(27, 26, 14.0777)
    assert abs(v.decimaldegrees - 27.437244) < 0.00001

def test_sexagesimal_decode():
    """
    Test we can decode a sexagesimal coordinate
    """
    v = APRSSexagesimal.decode("02726.23-")
    assert v.ambiguity == APRSPositionAmbiguity.NONE
    assert (v.degrees, v.minutes) == (-27, 26)
    assert abs(v.seconds - 13.8) < 0.001

def test_sexagesimal_decode_ambiguity1():
    """
    Test we can decode a sexagesimal coordinate with level 1 ambiguity
    """
    v = APRSSexagesimal.decode("02726.2 -")
    assert v.ambiguity == APRSPositionAmbiguity.TENTH_MINUTE
    assert (v.degrees, v.minutes) == (-27, 26)
    assert abs(v.seconds - 12.0) < 0.001

def test_sexagesimal_decode_ambiguity2():
    """
    Test we can decode a sexagesimal coordinate with level 2 ambiguity
    """
    v = APRSSexagesimal.decode("02726.  -")
    assert v.ambiguity == APRSPositionAmbiguity.MINUTE
    assert (v.degrees, v.minutes, v.seconds) == (-27, 26, 0)

def test_sexagesimal_decode_ambiguity3():
    """
    Test we can decode a sexagesimal coordinate with level 3 ambiguity
    """
    v = APRSSexagesimal.decode("0272 .  -")
    assert v.ambiguity == APRSPositionAmbiguity.TEN_MINUTES
    assert (v.degrees, v.minutes, v.seconds) == (-27, 20, 0)

def test_sexagesimal_decode_ambiguity4():
    """
    Test we can decode a sexagesimal coordinate with level 4 ambiguity
    """
    v = APRSSexagesimal.decode("027  .  -")
    assert v.ambiguity == APRSPositionAmbiguity.DEGREE
    assert (v.degrees, v.minutes, v.seconds) == (-27, 0, 0)

def test_sexagesimal_encode():
    """
    Test we can encode a sexagesimal coordinate
    """
    v = APRSSexagesimal(-27.437244)
    assert str(v) == "02726.23-"

def test_sexagesimal_decode_ambiguity1():
    """
    Test we can encode a sexagesimal coordinate with level 1 ambiguity
    """
    v = APRSSexagesimal(
            -27.437244, ambiguity=APRSPositionAmbiguity.TENTH_MINUTE
    )
    assert str(v) == "02726.2 -"

def test_sexagesimal_encode_ambiguity2():
    """
    Test we can encode a sexagesimal coordinate with level 2 ambiguity
    """
    v = APRSSexagesimal(
            -27.437244, ambiguity=APRSPositionAmbiguity.MINUTE
    )
    assert str(v) == "02726.  -"

def test_sexagesimal_encode_ambiguity3():
    """
    Test we can encode a sexagesimal coordinate with level 3 ambiguity
    """
    v = APRSSexagesimal(
            -27.437244, ambiguity=APRSPositionAmbiguity.TEN_MINUTES
    )
    assert str(v) == "0272 .  -"

def test_sexagesimal_encode_ambiguity4():
    """
    Test we can encode a sexagesimal coordinate with level 4 ambiguity
    """
    v = APRSSexagesimal(-27.437244, ambiguity=APRSPositionAmbiguity.DEGREE)
    assert str(v) == "027  .  -"

def test_sexagesimal_decode_tooshort():
    """
    Test the decoder refuses to decode a "too short" string.
    """
    try:
        APRSSexagesimal.decode("27")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Position string too short"

def test_sexagesimal_decode_unrecognised_sign():
    """
    Test the decoder refuses to decode an unknown sign character
    """
    try:
        APRSSexagesimal.decode("02726.25X")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Unrecognised sign: 'X'"

def test_sexagesimal_decode_nodot():
    """
    Test the decoder refuses to decode a string lacking a decimal point
    """
    try:
        APRSSexagesimal.decode("02726025+")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "No decimal point found in required position"

def test_sexagesimal_decimal_degrees_minutes():
    """
    Test the constructor refuses minutes with decimal degrees
    """
    try:
        APRSSexagesimal(-27.437244, minutes=30)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "Cannot define minutes or seconds when "
                "given decimal degrees"
        )

def test_sexagesimal_decimal_degrees_seconds():
    """
    Test the constructor refuses minutes with decimal degrees
    """
    try:
        APRSSexagesimal(-27.437244, seconds=30)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "Cannot define minutes or seconds when "
                "given decimal degrees"
        )

def test_sexagesimal_minutes_negative():
    """
    Test the constructor refuses negative minutes
    """
    try:
        APRSSexagesimal(-27, minutes=-26, seconds=30)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "minutes must be in the range [0..59]"
        )

def test_sexagesimal_minutes_toobig():
    """
    Test the constructor refuses too many minutes
    """
    try:
        APRSSexagesimal(-27, minutes=61, seconds=30)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "minutes must be in the range [0..59]"
        )

def test_sexagesimal_decimal_minutes_seconds():
    """
    Test the constructor refuses seconds with decimal minutes
    """
    try:
        APRSSexagesimal(-27, minutes=26.25, seconds=30)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "Cannot define seconds when "
                "given decimal minutes"
        )

def test_sexagesimal_seconds_negative():
    """
    Test the constructor refuses negative seconds
    """
    try:
        APRSSexagesimal(-27, seconds=-26)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "seconds must be in the range [0..59]"
        )

def test_sexagesimal_seconds_toobig():
    """
    Test the constructor refuses too many seconds
    """
    try:
        APRSSexagesimal(-27, seconds=61)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "seconds must be in the range [0..59]"
        )

def test_uncompressed_decode_tooshort():
    """
    Test the uncompressed coordinate decoder rejects too-short strings
    """
    try:
        APRSUncompressedCoordinates.decode("2724.35S/12345.67E")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == (
                "Co-ordinate string is too short"
        )

def test_uncompressed_decode():
    """
    Test the uncompressed coordinate can decode a valid string
    """
    coord = APRSUncompressedCoordinates.decode("2724.35S/12345.67E$")
    assert coord.lat.degrees == -27
    assert coord.lat.minutes == 24
    assert abs(coord.lat.seconds - 21.0) < 0.001
    assert coord.lng.degrees == 123
    assert coord.lng.minutes == 45
    assert abs(coord.lng.seconds - 40.2) < 0.001
    assert coord.symbol.tableident == "/"
    assert coord.symbol.symbol == "$"

def test_uncompressed_encode():
    """
    Test we can encode a valid uncompressed co-ordinate string
    """
    coord = APRSUncompressedCoordinates(
            lat=APRSLatitude(-27, 26.2354),
            lng=APRSLongitude(152, 56.5593),
            symbol=APRSSymbol("/", "$")
    )
    assert str(coord) == "2726.23S/15256.55E$"

def test_longitude_compressed_encode():
    """
    Test compressed longitude encoding
    """
    # From the APRS 1.0 protocol specs
    assert str(APRSCompressedLongitude(-72.75)) == "<*e7"

def test_longitude_compressed_decode():
    """
    Test compressed longitude decoding
    """
    # From the APRS 1.0 protocol specs
    v = APRSCompressedLongitude.decode("<*e7")
    assert (v.degrees, v.minutes) == (-72, 45)
    assert v.seconds < 0.02

def test_longitude_compressed_decode_toolong():
    """
    Test compressed longitude decoding rejects too short string
    """
    try:
        APRSCompressedLongitude.decode("<*e")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Compressed co-ordinate too short"

def test_csr_too_short():
    """
    Test a too-short course/speed/range/altitude is rejected
    """
    try:
        APRSCompressedCourseSpeedRange.decode("X", None)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Course/Speed value too short"

def test_latitude_compressed_encode():
    """
    Test compressed latitude encoding
    """
    # From the APRS 1.0 protocol specs
    assert str(APRSCompressedLatitude(49.50)) == "5L!!"

def test_latitude_compressed_decode():
    """
    Test compressed latitude decoding
    """
    # From the APRS 1.0 protocol specs
    v = APRSCompressedLatitude.decode("5L!!")
    assert (v.degrees, v.minutes) == (49, 30)
    assert v.seconds < 0.02

def test_latitude_compressed_decode_toolong():
    """
    Test compressed latitude decoding rejects too short string
    """
    try:
        APRSCompressedLatitude.decode("<*e")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Compressed co-ordinate too short"

def test_csr_too_short():
    """
    Test a too-short course/speed/range/altitude is rejected
    """
    try:
        APRSCompressedCourseSpeedRange.decode("X", None)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Course/Speed value too short"

def test_compressed_decode_too_short():
    try:
        APRSCompressedCoordinates.decode("X")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Co-ordinate too short for compressed format"

def test_compressed_decode_nocst():
    coord = APRSCompressedCoordinates.decode("/5L!!<*e7> sT")
    assert coord.lat.degrees == 49
    assert coord.lat.minutes == 30
    assert abs(coord.lat.seconds) < 0.001
    assert coord.lng.degrees == -72
    assert coord.lng.minutes == 45
    assert abs(coord.lng.seconds) < 0.02
    assert coord.symbol.tableident == "/"
    assert coord.symbol.symbol == ">"
    assert coord.ctype is None
    assert coord.csr is None

def test_compressed_decode_cs():
    coord = APRSCompressedCoordinates.decode("/5L!!<*e7>7P!")
    assert coord.lat.degrees == 49
    assert coord.lat.minutes == 30
    assert abs(coord.lat.seconds) < 0.001
    assert coord.lng.degrees == -72
    assert coord.lng.minutes == 45
    assert abs(coord.lng.seconds) < 0.02
    assert coord.symbol.tableident == "/"
    assert coord.symbol.symbol == ">"
    assert coord.ctype.gpsfix == APRSCompressionTypeGPSFix.OLD
    assert coord.ctype.nmeasrc == APRSCompressionTypeNMEASrc.OTHER
    assert coord.ctype.origin == APRSCompressionTypeOrigin.COMPRESSED
    assert coord.csr.altitude is None
    assert abs(coord.csr.speed - 36.2) < 0.1
    assert coord.csr.course == 88
    assert coord.csr.rng is None

def test_compressed_decode_rng():
    coord = APRSCompressedCoordinates.decode("/5L!!<*e7>{?!")
    assert coord.lat.degrees == 49
    assert coord.lat.minutes == 30
    assert abs(coord.lat.seconds) < 0.001
    assert coord.lng.degrees == -72
    assert coord.lng.minutes == 45
    assert abs(coord.lng.seconds) < 0.02
    assert coord.symbol.tableident == "/"
    assert coord.symbol.symbol == ">"
    assert coord.ctype.gpsfix == APRSCompressionTypeGPSFix.OLD
    assert coord.ctype.nmeasrc == APRSCompressionTypeNMEASrc.OTHER
    assert coord.ctype.origin == APRSCompressionTypeOrigin.COMPRESSED
    assert coord.csr.altitude is None
    assert coord.csr.speed is None
    assert coord.csr.course is None
    assert abs(coord.csr.rng - 20.125) < 0.01

def test_compressed_decode_alt():
    coord = APRSCompressedCoordinates.decode("/5L!!<*e7>S]1")
    assert coord.lat.degrees == 49
    assert coord.lat.minutes == 30
    assert abs(coord.lat.seconds) < 0.001
    assert coord.lng.degrees == -72
    assert coord.lng.minutes == 45
    assert abs(coord.lng.seconds) < 0.02
    assert coord.symbol.tableident == "/"
    assert coord.symbol.symbol == ">"
    assert coord.ctype.gpsfix == APRSCompressionTypeGPSFix.OLD
    assert coord.ctype.nmeasrc == APRSCompressionTypeNMEASrc.GGA
    assert coord.ctype.origin == APRSCompressionTypeOrigin.COMPRESSED
    assert abs(coord.csr.altitude - 10004.520) < 0.001
    assert coord.csr.speed is None
    assert coord.csr.course is None
    assert coord.csr.rng is None

def test_compressed_encode_nocst():
    assert str(APRSCompressedCoordinates(
            lat=APRSCompressedLatitude(-27.123),
            lng=APRSCompressedLongitude(152.53),
            symbol=APRSSymbol("/", "$"),
            ctype=None,
            csr=None
    )) == "/\\3\\pu%2;$ sT"

def test_compressed_encode_cs():
    assert str(APRSCompressedCoordinates(
            lat=APRSCompressedLatitude(-27.123),
            lng=APRSCompressedLongitude(152.53),
            symbol=APRSSymbol("/", "$"),
            ctype=APRSCompressionType(
                gpsfix=APRSCompressionTypeGPSFix.OLD,
                nmeasrc=APRSCompressionTypeNMEASrc.OTHER,
                origin=APRSCompressionTypeOrigin.COMPRESSED,
            ),
            csr=APRSCompressedCourseSpeedRange(
                course=88, speed=36.2
            )
    )) == "/\\3\\pu%2;$7O!"

def test_compressed_encode_gpsfix_current():
    assert str(APRSCompressedCoordinates(
            lat=APRSCompressedLatitude(-27.123),
            lng=APRSCompressedLongitude(152.53),
            symbol=APRSSymbol("/", "$"),
            ctype=APRSCompressionType(
                gpsfix=APRSCompressionTypeGPSFix.CURRENT,
                nmeasrc=APRSCompressionTypeNMEASrc.OTHER,
                origin=APRSCompressionTypeOrigin.COMPRESSED,
            ),
            csr=APRSCompressedCourseSpeedRange(
                course=88, speed=36.2
            )
    )) == "/\\3\\pu%2;$7OA"

def test_compressed_encode_nmeasrc_gga():
    assert str(APRSCompressedCoordinates(
            lat=APRSCompressedLatitude(-27.123),
            lng=APRSCompressedLongitude(152.53),
            symbol=APRSSymbol("/", "$"),
            ctype=APRSCompressionType(
                gpsfix=APRSCompressionTypeGPSFix.OLD,
                nmeasrc=APRSCompressionTypeNMEASrc.GGA,
                origin=APRSCompressionTypeOrigin.COMPRESSED,
            ),
            csr=APRSCompressedCourseSpeedRange(
                course=88, speed=36.2
            )
    )) == "/\\3\\pu%2;$7O1"

def test_compressed_encode_origin_software():
    assert str(APRSCompressedCoordinates(
            lat=APRSCompressedLatitude(-27.123),
            lng=APRSCompressedLongitude(152.53),
            symbol=APRSSymbol("/", "$"),
            ctype=APRSCompressionType(
                gpsfix=APRSCompressionTypeGPSFix.OLD,
                nmeasrc=APRSCompressionTypeNMEASrc.OTHER,
                origin=APRSCompressionTypeOrigin.SOFTWARE,
            ),
            csr=APRSCompressedCourseSpeedRange(
                course=88, speed=36.2
            )
    )) == "/\\3\\pu%2;$7O#"

def test_compressed_encode_rng():
    assert str(APRSCompressedCoordinates(
            lat=APRSCompressedLatitude(-27.123),
            lng=APRSCompressedLongitude(152.53),
            symbol=APRSSymbol("/", "$"),
            ctype=APRSCompressionType(
                gpsfix=APRSCompressionTypeGPSFix.OLD,
                nmeasrc=APRSCompressionTypeNMEASrc.OTHER,
                origin=APRSCompressionTypeOrigin.COMPRESSED,
            ),
            csr=APRSCompressedCourseSpeedRange(
                rng=20.125
            )
    )) == "/\\3\\pu%2;${>!"

def test_compressed_encode_alt():
    assert str(APRSCompressedCoordinates(
            lat=APRSCompressedLatitude(-27.123),
            lng=APRSCompressedLongitude(152.53),
            symbol=APRSSymbol("/", "$"),
            ctype=APRSCompressionType(
                gpsfix=APRSCompressionTypeGPSFix.OLD,
                nmeasrc=APRSCompressionTypeNMEASrc.OTHER,
                origin=APRSCompressionTypeOrigin.COMPRESSED,
            ),
            csr=APRSCompressedCourseSpeedRange(
                altitude=10004.520
            )
    )) == "/\\3\\pu%2;$S\\!"

def test_compressed_decode_invalid_reject():
    try:
        APRSCompressedCoordinates.decode("/5L!!<*e7>~~!")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Unknown Course/Speed/Range field: [93, 93]"

def test_compressed_construct_csr_noctype():
    try:
        APRSCompressedCoordinates(
                lat=APRSCompressedLatitude(-27.123),
                lng=APRSCompressedLongitude(152.53),
                symbol=APRSSymbol("/", "$"),
                csr=APRSCompressedCourseSpeedRange(course=50, speed=10)
        )
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "If either ctype or csr is given, give both"

def test_compressed_construct_noccsr_type():
    try:
        APRSCompressedCoordinates(
                lat=APRSCompressedLatitude(-27.123),
                lng=APRSCompressedLongitude(152.53),
                symbol=APRSSymbol("/", "$"),
                ctype=APRSCompressionType(
                    gpsfix=APRSCompressionTypeGPSFix.OLD,
                    nmeasrc=APRSCompressionTypeNMEASrc.OTHER,
                    origin=APRSCompressionTypeOrigin.COMPRESSED
                )
        )
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "If either ctype or csr is given, give both"

def test_csr_construct_alt_course():
    try:
        APRSCompressedCourseSpeedRange(altitude=5000, course=50)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ( 
                "Altitude cannot be specified with " 
                "range, course or speed" 
        )

def test_csr_construct_alt_speed():
    try:
        APRSCompressedCourseSpeedRange(altitude=5000, speed=50)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ( 
                "Altitude cannot be specified with " 
                "range, course or speed" 
        )

def test_csr_construct_alt_rng():
    try:
        APRSCompressedCourseSpeedRange(altitude=5000, rng=50)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ( 
                "Altitude cannot be specified with " 
                "range, course or speed" 
        )

def test_csr_construct_rng_course():
    try:
        APRSCompressedCourseSpeedRange(rng=100, course=50)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ( 
                "Range cannot be specified with " 
                "altitude, course or speed" 
        )

def test_csr_construct_rng_speed():
    try:
        APRSCompressedCourseSpeedRange(rng=100, speed=50)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ( 
                "Range cannot be specified with " 
                "altitude, course or speed" 
        )

def test_csr_construct_course_nospeed():
    try:
        APRSCompressedCourseSpeedRange(course=92)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ( 
                "Course and speed must both be specified" 
        )

def test_csr_construct_speed_nocourse():
    try:
        APRSCompressedCourseSpeedRange(speed=92)
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ( 
                "Course and speed must both be specified" 
        )

def test_csr_get_speed_q():
    """
    Test we can obtain the speed as a Quantity
    """
    csr = APRSCompressedCourseSpeedRange(speed=20, course=50)
    speed = csr.speed_q
    assert isinstance(speed, Quantity)
    assert speed.magnitude == 20.0
    assert str(speed.units) == "knot"

def test_csr_set_speed_q():
    """
    Test we can update the speed as a Quantity
    """
    csr = APRSCompressedCourseSpeedRange(speed=20, course=50)
    csr.speed_q = Quantity(45, "knot")
    assert csr.speed == 45

def test_csr_get_rng_q():
    """
    Test we can obtain the range as a Quantity
    """
    csr = APRSCompressedCourseSpeedRange(rng=30)
    rng = csr.rng_q
    assert isinstance(rng, Quantity)
    assert rng.magnitude == 30.0
    assert str(rng.units) == "mile"

def test_csr_set_rng_q():
    """
    Test we can update the range as a Quantity
    """
    csr = APRSCompressedCourseSpeedRange(rng=30)
    csr.rng_q = Quantity(60, "mile")
    assert csr.rng == 60

def test_csr_get_altitude_q():
    """
    Test we can obtain the altitude as a Quantity
    """
    csr = APRSCompressedCourseSpeedRange(altitude=8000)
    altitude = csr.altitude_q
    assert isinstance(altitude, Quantity)
    assert altitude.magnitude == 8000.0
    assert str(altitude.units) == "foot"

def test_csr_set_altitude_q():
    """
    Test we can update the altitude as a Quantity
    """
    csr = APRSCompressedCourseSpeedRange(altitude=8000)
    csr.altitude_q = Quantity(6000, "foot")
    assert csr.altitude == 6000

def test_decode_invalid_type():
    """
    Test the decode routine rejects unknown APRS message types
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'X3722.20N/07900.66W&000/000/A=000685Mobile'
    )
    payload = frame.payload.decode("US-ASCII")
    try:
        APRSPositionFrame.decode(
                frame, payload,
                logging.getLogger('decoder')
        )
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ("Not a recognised frame: %r" % payload)

def test_decode_not_pos_type():
    """
    Test the decode routine rejects non-position-reports
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b':3722.20N/07900.66W&000/000/A=000685Mobile'
    )
    payload = frame.payload.decode("US-ASCII")
    try:
        APRSPositionFrame.decode(
                frame, payload,
                logging.getLogger('decoder')
        )
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == ("Not a position frame: %r" % payload)

def test_decode_position_nocomment():
    """
    Test decode can decode a position report without comment.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'!3722.20N/07900.66W&'
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert not decoded.has_messaging
    assert decoded.position_ts is None
    assert decoded.position.lat.degrees == 37
    assert decoded.position.lat.minutes == 22
    assert abs(decoded.position.lat.seconds - 12) < 0.001
    assert decoded.position.lng.degrees == -79
    assert decoded.position.lng.minutes == 0
    assert abs(decoded.position.lng.seconds - 39.6) < 0.001
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == "&"
    assert decoded.message is None

def test_decode_uncompressed_position_nots():
    """
    Test decode can decode an uncompressed position report without timestamp.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'!3722.20N/07900.66W&000/000/A=000685Mobile'
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert not decoded.has_messaging
    assert decoded.position_ts is None
    assert decoded.position.lat.degrees == 37
    assert decoded.position.lat.minutes == 22
    assert abs(decoded.position.lat.seconds - 12) < 0.001
    assert decoded.position.lng.degrees == -79
    assert decoded.position.lng.minutes == 0
    assert abs(decoded.position.lng.seconds - 39.6) < 0.001
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == "&"
    assert decoded.message == \
        "000/000/A=000685Mobile"

def test_decode_uncompressed_position_nots_msgcap():
    """
    Test decode can decode an uncompressed position report
    advertising message capability
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'=3722.20N/07900.66W&000/000/A=000685Mobile'
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert decoded.has_messaging
    assert decoded.position_ts is None
    assert decoded.position.lat.degrees == 37
    assert decoded.position.lat.minutes == 22
    assert abs(decoded.position.lat.seconds - 12) < 0.001
    assert decoded.position.lng.degrees == -79
    assert decoded.position.lng.minutes == 0
    assert abs(decoded.position.lng.seconds - 39.6) < 0.001
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == "&"
    assert decoded.message == \
        "000/000/A=000685Mobile"

def test_decode_uncompressed_position_withts():
    """
    Test decode can decode an uncompressed position report with timestamp.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'/092345z3722.20N/07900.66W&000/000/A=000685Mobile'
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert not decoded.has_messaging
    assert decoded.position_ts.day == 9
    assert decoded.position_ts.hour == 23
    assert decoded.position_ts.minute == 45
    assert decoded.position.lat.degrees == 37
    assert decoded.position.lat.minutes == 22
    assert abs(decoded.position.lat.seconds - 12) < 0.001
    assert decoded.position.lng.degrees == -79
    assert decoded.position.lng.minutes == 0
    assert abs(decoded.position.lng.seconds - 39.6) < 0.001
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == "&"
    assert decoded.message == \
        "000/000/A=000685Mobile"

def test_decode_uncompressed_position_withts_withmsgcap():
    """
    Test decode can decode an uncompressed position report with timestamp
    advertising messaging capability
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'@092345z3722.20N/07900.66W&000/000/A=000685Mobile'
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert decoded.has_messaging
    assert decoded.position_ts.day == 9
    assert decoded.position_ts.hour == 23
    assert decoded.position_ts.minute == 45
    assert decoded.position.lat.degrees == 37
    assert decoded.position.lat.minutes == 22
    assert abs(decoded.position.lat.seconds - 12) < 0.001
    assert decoded.position.lng.degrees == -79
    assert decoded.position.lng.minutes == 0
    assert abs(decoded.position.lng.seconds - 39.6) < 0.001
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == "&"
    assert decoded.message == \
        "000/000/A=000685Mobile"

def test_decode_compressed_position_nots():
    """
    Test decode can decode an compressed position report without timestamp.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'!/5L!!<*e7& sT000/000/A=000685Mobile'
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert not decoded.has_messaging
    assert decoded.position_ts is None
    assert decoded.position.lat.degrees == 49
    assert decoded.position.lat.minutes == 30
    assert abs(decoded.position.lat.seconds) < 0.001
    assert decoded.position.lng.degrees == -72
    assert decoded.position.lng.minutes == 45
    assert abs(decoded.position.lng.seconds) < 0.02
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == "&"
    assert decoded.message == \
        "000/000/A=000685Mobile"

def test_decode_compressed_position_withts():
    """
    Test decode can decode an compressed position report with timestamp.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'/092345z/5L!!<*e7& sT000/000/A=000685Mobile'
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert not decoded.has_messaging
    assert decoded.position_ts.day == 9
    assert decoded.position_ts.hour == 23
    assert decoded.position_ts.minute == 45
    assert decoded.position.lat.degrees == 49
    assert decoded.position.lat.minutes == 30
    assert abs(decoded.position.lat.seconds) < 0.001
    assert decoded.position.lng.degrees == -72
    assert decoded.position.lng.minutes == 45
    assert abs(decoded.position.lng.seconds) < 0.02
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == "&"
    assert decoded.message == \
        "000/000/A=000685Mobile"

def test_decode_compressed_position_withcsr():
    """
    Test decode can decode an compressed position report with Course/Speed.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'=/5L!!<*e7>7P[' # APRS101.PDF page 41
    )
    decoded = APRSPositionFrame.decode(
            frame, frame.payload.decode("US-ASCII"),
            logging.getLogger('decoder')
    )
    assert isinstance(decoded, APRSPositionFrame)
    assert decoded.has_messaging
    assert decoded.position_ts is None
    print ("Latitude:  %3d° %2d' %6.3f\"" % (
        decoded.position.lat.degrees,
        decoded.position.lat.minutes,
        decoded.position.lat.seconds
    ))
    print ("Longitude: %3d° %2d' %6.3f\"" % (
        decoded.position.lng.degrees,
        decoded.position.lng.minutes,
        decoded.position.lng.seconds
    ))
    assert decoded.position.lat.degrees == 49
    assert decoded.position.lat.minutes == 30
    assert abs(decoded.position.lat.seconds) < 0.001
    assert decoded.position.lng.degrees == -72
    assert decoded.position.lng.minutes == 45
    assert abs(decoded.position.lng.seconds) < 0.02
    assert decoded.position.symbol.tableident == "/"
    assert decoded.position.symbol.symbol == ">"
    assert decoded.message is None
    assert decoded.position.ctype.gpsfix == APRSCompressionTypeGPSFix.CURRENT
    assert decoded.position.ctype.nmeasrc == APRSCompressionTypeNMEASrc.RMC
    assert decoded.position.ctype.origin == APRSCompressionTypeOrigin.SOFTWARE
    assert decoded.position.csr.course == 88
    assert abs(decoded.position.csr.speed - 36.2) < 0.1

def test_encode_position_nocomment():
    """
    Test we can encode a position report without comment.
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSUncompressedCoordinates(
                lat=APRSLatitude(-27, 26.2354),
                lng=APRSLongitude(152, 56.5593),
                symbol=APRSSymbol("/", "&")
            ),
            position_ts=None,
            message=None,
            messaging=False
    )
    assert frame.payload == b"!2726.23S/15256.55E&"

def test_encode_uncompressed_position_nots():
    """
    Test we can encode an uncompressed position report without timestamp.
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSUncompressedCoordinates(
                lat=APRSLatitude(-27, 26.2354),
                lng=APRSLongitude(152, 56.5593),
                symbol=APRSSymbol("/", "&")
            ),
            position_ts=None,
            message="Mobile",
            messaging=False
    )
    assert frame.payload == b"!2726.23S/15256.55E&Mobile"

def test_encode_uncompressed_position_nots_msgcap():
    """
    Test we can encode an uncompressed position report
    advertising message capability
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSUncompressedCoordinates(
                lat=APRSLatitude(-27, 26.2354),
                lng=APRSLongitude(152, 56.5593),
                symbol=APRSSymbol("/", "&")
            ),
            position_ts=None,
            message="Mobile",
            messaging=True
    )
    assert frame.payload == b"=2726.23S/15256.55E&Mobile"

def test_encode_uncompressed_position_withts():
    """
    Test decode can decode an uncompressed position report with timestamp.
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSUncompressedCoordinates(
                lat=APRSLatitude(-27, 26.2354),
                lng=APRSLongitude(152, 56.5593),
                symbol=APRSSymbol("/", "&")
            ),
            position_ts=MDHMTimestamp(7, 31, 6, 33),
            message=None,
            messaging=False
    )
    assert frame.payload == b"/073106332726.23S/15256.55E&"

def test_encode_uncompressed_position_withts_withmsgcap():
    """
    Test decode can decode an uncompressed position report with timestamp
    advertising messaging capability
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSUncompressedCoordinates(
                lat=APRSLatitude(-27, 26.2354),
                lng=APRSLongitude(152, 56.5593),
                symbol=APRSSymbol("/", "&")
            ),
            position_ts=MDHMTimestamp(7, 31, 6, 33),
            message="My Message",
            messaging=True
    )
    assert frame.payload == b"@073106332726.23S/15256.55E&My Message"

def test_encode_compressed_position_nots():
    """
    Test we can encode an compressed position report without timestamp.
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSCompressedCoordinates(
                lat=APRSCompressedLatitude(-27.437256),
                lng=APRSLongitude(152.942655),
                symbol=APRSSymbol("/", "&")
            ),
            position_ts=None,
            message=None,
            messaging=False
    )
    assert frame.payload == b"!/\\B+A15256.55E& sT"

def test_encode_compressed_position_withts():
    """
    Test decode can decode an compressed position report with timestamp.
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSCompressedCoordinates(
                lat=APRSCompressedLatitude(-27.437256),
                lng=APRSLongitude(152.942655),
                symbol=APRSSymbol("/", "&")
            ),
            position_ts=MDHMTimestamp(7, 31, 6, 33),
            message=None,
            messaging=False
    )
    assert frame.payload == b"/07310633/\\B+A15256.55E& sT"

def test_encode_compressed_position_withcsr():
    """
    Test we can encode an compressed position report with Course/Speed.
    """
    frame = APRSPositionFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            position=APRSCompressedCoordinates(
                lat=APRSCompressedLatitude(-27.437256),
                lng=APRSLongitude(152.942655),
                symbol=APRSSymbol("/", "&"),
                ctype=APRSCompressionType(
                    gpsfix=APRSCompressionTypeGPSFix.CURRENT,
                    nmeasrc=APRSCompressionTypeNMEASrc.GGA,
                    origin=APRSCompressionTypeOrigin.SOFTWARE
                ),
                csr=APRSCompressedCourseSpeedRange(
                    course=60, speed=32
                )
            ),
            position_ts=None,
            message=None,
            messaging=False
    )
    assert frame.payload == b"!/\\B+A15256.55E&0NS"



#!/usr/bin/env python3

"""
Dummy IOLoop interface
"""

import time

class DummyLoop(object):
    def __init__(self):
        self.readers = {}
        self.calls = []

    def time(self):
        return time.monotonic()

    def call_soon(self, callback, *args):
        self.calls.append((self.time(), callback) + args)

    def call_later(self, delay, callback, *args):
        when = self.time() + delay
        self.calls.append((when, callback) + args)
        return DummyTimeoutHandle(when)

    def add_reader(self, fileno, reader):
        self.readers[fileno] = reader

    def remove_reader(self, fileno):
        self.readers.pop(fileno)


class DummyTimeoutHandle(object):
    def __init__(self, when):
        self._when = when
        self._cancelled = False

    def cancel(self):
        self._cancelled = True

    # Note: Don't rely on this in real code in Python <3.6 as
    # these methods were added in 3.7.

    def when(self):
        return self._when

    def cancelled(self):
        return self._cancelled



#!/usr/bin/env python3

"""
Unit handling tests
"""

from pytest import skip

try:
    from pint import Quantity
    PINT_SUPPORTED = True
except ImportError:
    PINT_SUPPORTED = False

from aioax25.unit import checknumeric, convertvalue

def test_checknumeric_required_none():
    """
    checknumeric should raise ValueError if required value is None
    """
    try:
        checknumeric("myparam", None, required=True)
        assert False, "Should not have passed"
    except ValueError as e:
        assert str(e) == "myparam is a required parameter"


def test_checknumeric_optional_none():
    """
    checknumeric should return None if optional value is None
    """
    assert checknumeric("myparam", None, required=False) is None, \
            "Should have passed through None for optional parameter"


def test_checknumeric_int():
    """
    checknumeric should cast int to float
    """
    v = checknumeric("myparam", 1234)
    assert v == 1234.0
    assert isinstance(v, float), "Should cast to float"


def test_checknumeric_str():
    """
    checknumeric should cast str to float
    """
    v = checknumeric("myparam", "123.45")
    assert v == 123.45
    assert isinstance(v, float), "Should cast to float"


def test_checknumeric_float():
    """
    checknumeric should pass through float
    """
    v = checknumeric("myparam", 1234.56)
    assert v == 1234.56


def test_convertvalue_none():
    """
    convertvalue should handle None without barfing
    """
    assert convertvalue("optparam", None, "m", required=False) is None, \
            "Should have returned None here"

def test_convertvalue_barevalue():
    """
    convertvalue should pass through bare value
    """
    assert convertvalue("optparam", 123.45, "m", required=False) == 123.45, \
            "Should have passed through value"


def test_convertvalue_quantity():
    """
    convertvalue should convert Quantity to correct unit
    """
    if not PINT_SUPPORTED:
        skip(
            "pint.Quantity could not be imported, "
            "so unit conversion won't work as expected."
        )

    assert convertvalue("optparam", \
            Quantity(1, "in"), "cm", required=False) == 2.54, \
            "Should have converted the value"



#!/usr/bin/env python3

"""
A logging.Logger substitute for testing purposes.
"""

# Was going to use pytest_logdog, *BUT* that needs Python 3.7+ and I'm
# targetting Python 3.4!

import pytest
from sys import exc_info


class DummyLogger(object):
    def __init__(self):
        self.logrecords = []
        self.children = {}

    def _addrecord(self, log_method, log_args, log_kwargs, _exc_info=False):
        if _exc_info:
            (ex_type, ex_val, ex_tb) = exc_info()
        else:
            ex_type = None
            ex_val = None
            ex_tb = None

        self.logrecords.append(dict(
            method=log_method, args=log_args, kwargs=log_kwargs,
            ex_type=ex_type, ex_val=ex_val, ex_tb=ex_tb
        ))

    # Message logging endpoints
    def critical(self, *args, exc_info=False, **kwargs):
        self._addrecord('critical', args, kwargs, exc_info)

    def debug(self, *args, exc_info=False, **kwargs):
        self._addrecord('debug', args, kwargs, exc_info)

    def error(self, *args, exc_info=False, **kwargs):
        self._addrecord('error', args, kwargs, exc_info)

    def exception(self, *args, **kwargs):
        self._addrecord('exception', args, kwargs, True)

    def fatal(self, *args, exc_info=False, **kwargs):
        self._addrecord('fatal', args, kwargs, exc_info)

    def info(self, *args, exc_info=False, **kwargs):
        self._addrecord('info', args, kwargs, exc_info)

    def log(self, *args, exc_info=False, **kwargs):
        self._addrecord('log', args, kwargs, exc_info)

    def warn(self, *args, exc_info=False, **kwargs):
        self._addrecord('warn', args, kwargs, exc_info)

    def warning(self, *args, exc_info=False, **kwargs):
        self._addrecord('warning', args, kwargs, exc_info)

    # Info endpoints
    def isEnabledFor(self, level):
        return True

    # Hierarchy endpoints
    def getChild(self, suffix):
        try:
            return self.children[suffix]
        except KeyError:
            child = DummyLogger()
            self.children[suffix] = child
            return child


@pytest.fixture
def logger():
    """Dummy logger instance that mocks logging.Logger."""
    return DummyLogger()



#!/usr/bin/env python3

"""
Unit handling.  Optionally wraps `pint`, otherwise we just
assert numeric types and assume correct units.
"""

def checknumeric(name, value, required=False):
    """
    Assert the value is a numeric value, or throw a TypeError.
    """
    if value is None:
        if required:
            raise ValueError("%s is a required parameter" % name)
        else:
            return None

    # This will throw ValueError if it can't be converted
    return float(value)


# Optional dependency: pint (for unit conversion)
try:
    from pint import Quantity

    def convertvalue(name, quantity, units, required=False):
        """
        Assert the value is a numeric value and convert to the appropriate
        units if possible.
        """
        if (quantity is not None) and isinstance(quantity, Quantity):
            # Convert to target units, take the magnitude
            return quantity.to(units).magnitude
        else:
            # Pass through to handler
            return checknumeric(name, quantity, required=required)
except ImportError: # pragma: no cover
    class Quantity(object):
        """
        Dummy Quantity class work-alike for systems without `pint`.
        """
        def __init__(self, magnitude, units):
            self.magnitude = magnitude
            self.units = units

        def __str__(self):
            return "%r %s" % (self.magnitude, self.units)

        def __repr__(self):
            return "Quantity(magnitude=%r, units=%r)" \
                    % (self.magnitude, self.units)

        def to(self, units, *a, **kwa):
            if units == self.units:
                # No conversion required
                return self

            raise NotImplementedError(
                "Unit conversion is not implemented here.  "\
                "`pip install pint` if you want unit conversion."
            )

    def convertvalue(name, quantity, units, required=False):
        if (quantity is not None) and isinstance(quantity, Quantity):
            # Assert correct units!
            if quantity.units != units:
                raise ValueError(
                        "%s parameter must be in %s units (`pip install pint` " 
                        "for unit conversion)"\
                                % (name, units)
                )
            return quantity.magnitude
        else:
            # Pass through to handler
            return checknumeric(name, quantity, required=required)
